######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""QWidget that is used to create or edit Tool specifications.
In the former case it is presented empty, but in the latter it
is filled with all the information from the specification being edited."""
from copy import deepcopy
from enum import IntEnum, unique
import logging
from operator import methodcaller
import os
from PySide6.QtCore import QFileInfo, QItemSelection, QItemSelectionModel, QModelIndex, Qt, Slot
from PySide6.QtGui import QFont, QStandardItem, QStandardItemModel, QTextDocument
from PySide6.QtWidgets import QFileDialog, QFileIconProvider, QInputDialog, QLabel, QMessageBox
from spine_engine.utils.command_line_arguments import split_cmdline_args
from spine_items.tool.widgets.tool_spec_optional_widgets import (
    ExecutableToolSpecOptionalWidget,
    JuliaToolSpecOptionalWidget,
    PythonToolSpecOptionalWidget,
)
from spinetoolbox.config import STATUSBAR_SS
from spinetoolbox.helpers import SealCommand, busy_effect, open_url, same_path
from spinetoolbox.project_item.specification_editor_window import (
    ChangeSpecPropertyCommand,
    SpecificationEditorWindowBase,
    UniqueCommandId,
)
from spinetoolbox.widgets.custom_qwidgets import ToolBarWidget
from ..item_info import ItemInfo
from ..tool_specifications import TOOL_TYPES, make_specification


@unique
class CommandId(IntEnum):
    ARGS_UPDATE = UniqueCommandId.unique_id()
    EXECUTABLE_UPDATE = UniqueCommandId.unique_id()
    COMMAND_UPDATE = UniqueCommandId.unique_id()


class ToolSpecificationEditorWindow(SpecificationEditorWindowBase):
    """A widget to query user's preferences for a new tool specification."""

    def __init__(self, toolbox, specification=None, item=None):
        """
        Args:
            toolbox (ToolboxUI): QMainWindow instance
            specification (ToolSpecification, optional): If given, the form is pre-filled with this specification
            item (ProjectItem, optional): Sets the spec for this item if accepted
        """
        super().__init__(toolbox, specification, item)  # Inherit stylesheet from ToolboxUI
        self._project = self._toolbox.project()
        # Customize text edit main program
        self._ui.textEdit_program.setEnabled(False)
        self._ui.textEdit_program.file_selected(False)
        self._programfile_documents = {}
        self._programfile_set_dirty_slots = {}
        # Setup statusbar
        self._label_main_path = QLabel()
        label = QLabel("Source dir:")
        font = QFont()
        font.setPointSize(10)
        label.setFont(font)
        font = QFont(font)
        font.setBold(True)
        self._label_main_path.setFont(font)
        self._ui.statusbar.addPermanentWidget(label)
        self._ui.statusbar.addPermanentWidget(self._label_main_path)
        self._ui.statusbar.setStyleSheet(STATUSBAR_SS)
        # init models
        self.programfiles_model = QStandardItemModel(self)
        self.io_files_model = QStandardItemModel(self)
        # init ui
        self._ui.treeView_programfiles.setModel(self.programfiles_model)
        self._ui.treeView_io_files.setModel(self.io_files_model)
        self._ui.comboBox_tooltype.addItems(TOOL_TYPES)
        self._ui.comboBox_tooltype.setCurrentIndex(-1)
        self.optional_widget = None
        # if a specification is given, fill the form with data from it
        if specification is not None:
            self._ui.lineEdit_args.setText(" ".join(specification.cmdline_args))
            tooltype = specification.tooltype.lower()
            index = next(iter(k for k, t in enumerate(TOOL_TYPES) if t.lower() == tooltype), -1)
            self._ui.comboBox_tooltype.setCurrentIndex(index)
            self._ui.textEdit_program.set_lexer_name(tooltype.lower())
            specification.init_execution_settings()  # Initialize execution settings
            # spec dict needs to be set after setting the execution settings but before the
            # optional widget is shown, in case the tooltype is Executable.
            self.spec_dict = deepcopy(specification.to_dict())
            self._show_optional_widget(index)
            self._init_optional_widget(specification)
        else:
            # If no specification is given, specification dict needs to be prepared
            self.spec_dict = {"item_type": ItemInfo.item_type()}
        # Init lists
        programfiles = list(specification.includes) if specification else []
        # Get first item from programfiles list as the main program file
        main_program_file = programfiles.pop(0) if programfiles else None
        inputfiles = list(specification.inputfiles) if specification else []
        inputfiles_opt = list(specification.inputfiles_opt) if specification else []
        outputfiles = list(specification.outputfiles) if specification else []
        self.includes_main_path = specification.path if specification else None
        if self.has_root_directory():
            self.includes_main_path = self.item.root_dir
        # Populate lists (this will also create headers)
        self.init_programfile_list()
        self.init_io_file_list()
        self.populate_programfile_list(programfiles)
        self.populate_inputfiles_list(inputfiles)
        self.populate_inputfiles_opt_list(inputfiles_opt)
        self.populate_outputfiles_list(outputfiles)
        if main_program_file:
            try:
                if self.has_root_directory():
                    self._set_main_program_file(os.path.abspath(os.path.join(self.item.root_dir, main_program_file)))
                else:
                    self._set_main_program_file(os.path.join(self.includes_main_path, main_program_file))
            except TypeError:
                self._set_main_program_file("")
        else:
            self._label_main_path.setText(self.includes_main_path)
            self._clear_program_text_edit()
        self.connect_signals()
        # Select main program file index
        parent = self.programfiles_model.index(0, 0)
        if self.programfiles_model.rowCount(parent) == 0:
            return
        index = self.programfiles_model.index(0, 0, parent)
        selection_model = self._ui.treeView_programfiles.selectionModel()
        selection_model.setCurrentIndex(index, QItemSelectionModel.SelectionFlag.Select)
        self._load_programfile_in_editor(index)
        self._enable_additional_program_files_actions()

    def _make_ui(self):
        from ..ui.tool_specification_form import Ui_MainWindow  # pylint: disable=import-outside-toplevel

        return Ui_MainWindow()

    def has_root_directory(self):
        if self.item is not None and self.item.root_dir != "":
            return True
        return False

    @property
    def settings_group(self):
        return "toolSpecificationEditorWindow"

    def _restore_dock_widgets(self):
        docks = (self._ui.dockWidget_program_files, self._ui.dockWidget_program)
        self.splitDockWidget(*docks, Qt.Orientation.Horizontal)
        width = sum(d.width() for d in docks)
        self.resizeDocks(docks, [width * x for x in (0.3, 0.7)], Qt.Orientation.Horizontal)
        docks = (self._ui.dockWidget_program_files, self._ui.dockWidget_io_files)
        self.splitDockWidget(*docks, Qt.Orientation.Vertical)
        height = sum(d.height() for d in docks)
        self.resizeDocks(docks, [height * x for x in (0.6, 0.4)], Qt.Orientation.Vertical)

    @Slot(bool)
    def _update_window_modified(self, _clean):
        clean = self._undo_stack.isClean() and not any(doc.isModified() for doc in self._programfile_documents.values())
        super()._update_window_modified(clean)

    def _make_optional_widget(self, toolspectype):
        """Returns an optional widget based on tool spec type.
        Embedded to the main window in ``self._show_optional_widget()``.

        Args:
            toolspectype (str): Tool spec type

        Returns:
            Union[ToolSpecOptionalWidget, None]
        """
        if toolspectype.lower() == "python":
            return PythonToolSpecOptionalWidget(self, self._toolbox)
        if toolspectype.lower() == "julia":
            return JuliaToolSpecOptionalWidget(self, self._toolbox)
        if toolspectype.lower() == "executable":
            return ExecutableToolSpecOptionalWidget(self, self._toolbox)
        return None

    def _get_optional_widget(self, toolspectype):
        """Returns the current optional widget for given tool spec type or None if unavailable.

        Args:
            toolspectype (str): Tool spec type

        Returns:
            Union[ToolSpecOptionalWidget, None]
        """
        if toolspectype.lower() in ("python", "julia", "executable"):
            optional_widget = self._ui.horizontalLayout_options_placeholder.itemAt(0)
            return optional_widget.widget()
        return None

    @Slot(int)
    def _show_optional_widget(self, row):
        """Adds a ToolSpecOptionalWidget to UI and updates the widget based on specification.

        Args:
            row (int): Active row in tool type combobox
        """
        optional_widget = self._ui.horizontalLayout_options_placeholder.takeAt(0)
        if optional_widget is not None:
            optional_widget.widget().close()
            self.optional_widget = None
            self._set_tab_order()
        tooltype = self._ui.comboBox_tooltype.itemText(row)
        self.optional_widget = self._make_optional_widget(tooltype)
        if self.optional_widget:
            self._ui.horizontalLayout_options_placeholder.addWidget(self.optional_widget)
            self.optional_widget.show()
            self._set_tab_order()
        if self.spec_dict.get("tooltype") == "executable":
            self.optional_widget.set_command_and_shell_edit_disabled_state(bool(self._current_main_program_file()))

    def _init_optional_widget(self, specification):
        """Initializes optional widget UI based on specification.

        Args:
            specification (ToolSpecification): Specification that is opened in the editor window
        """
        toolspectype = specification.tooltype.lower()
        opt_widget = self._get_optional_widget(toolspectype)
        if opt_widget:
            opt_widget.init_widget(specification)

    def _set_tab_order(self):
        self.setTabOrder(self._ui.comboBox_tooltype, self._ui.lineEdit_args)
        before = self._ui.lineEdit_args
        if self.optional_widget is not None:
            for after in self.optional_widget.get_widgets_in_tab_order():
                self.setTabOrder(before, after)
                before = after
        self.setTabOrder(before, self._ui.treeView_programfiles)
        self.setTabOrder(self._ui.treeView_programfiles, self._ui.treeView_io_files)
        self.setTabOrder(self._ui.treeView_io_files, self._ui.textEdit_program)

    def _make_new_specification(self, spec_name):
        """See base class."""
        if self._ui.comboBox_tooltype.currentIndex() == -1:  # Check that a tool type is selected
            self.show_error("Tool spec type not selected")
            return None
        toolspectype = self._ui.comboBox_tooltype.currentText().lower()
        # Check that Tool Spec has a main program file except for Executable Tool Specs
        main_program_file = self._current_main_program_file()
        if main_program_file is None and toolspectype != "executable":
            self.show_error("Please add a main program file")
            return None
        new_spec_dict = {"name": spec_name, "description": self._spec_toolbar.description(), "tooltype": toolspectype}
        if not main_program_file:
            main_prgm_file_name = ""
            if self.has_root_directory():
                self.includes_main_path = self.item.root_dir
            else:
                self.includes_main_path = None
        elif self.has_root_directory():
            main_prgm_file_name = os.path.relpath(main_program_file, self.item.root_dir)
            self.includes_main_path = self.item.root_dir
        else:
            main_prgm_dir, main_prgm_file_name = os.path.split(main_program_file)
            self.includes_main_path = os.path.abspath(main_prgm_dir)
        self._label_main_path.setText(self.includes_main_path)
        new_spec_dict["includes"] = [main_prgm_file_name] if main_prgm_file_name else []
        new_spec_dict["includes"] += self._additional_program_file_list()
        new_spec_dict["inputfiles"] = self._input_file_list()
        new_spec_dict["inputfiles_opt"] = self._opt_input_file_list()
        new_spec_dict["outputfiles"] = self._output_file_list()
        # Strip whitespace from args before saving it to JSON
        new_spec_dict["cmdline_args"] = split_cmdline_args(self._ui.lineEdit_args.text())
        new_spec_dict["includes_main_path"] = (
            self.includes_main_path.replace(os.sep, "/") if self.includes_main_path else None
        )
        try:
            new_spec_dict = self.insert_execution_settings(new_spec_dict)
        except NameError:
            self.show_error(f"Creating a {self._ui.comboBox_tooltype.currentText().lower()} Tool spec failed")
            return None
        tool_spec = make_specification(new_spec_dict, self._toolbox.qsettings(), self._toolbox)
        if not tool_spec:
            self.show_error("Creating Tool specification failed")
            return None
        return tool_spec

    def insert_execution_settings(self, d):
        """Insert execution mode settings to Tool spec dictionary.
        For Python Tool specs, these settings give the kernel spec name, is_conda boolean
        and whether to use Jupyter Console or not.

        Args:
            d (dict): Tool spec dictionary where the extra settings (if available) are included
        """
        toolspectype = d["tooltype"]
        opt_widget = self._get_optional_widget(toolspectype)
        if not opt_widget:
            return d
        d["execution_settings"] = opt_widget.add_execution_settings(toolspectype)
        return d

    def _save(self, exiting=None):
        """Saves spec. If successful, also saves all modified program files."""
        if not super()._save(exiting):
            return False
        saved = []
        for file_path, doc in self._programfile_documents.items():
            if not doc.isModified():
                continue
            err = self._save_program_file(file_path, doc)
            basename = os.path.basename(file_path)
            if err is None:
                doc.setModified(False)
                saved.append(basename)
                continue
            self.show_error(f"Error while saving {basename}: {err}")
            return True
        if saved:
            saved = ", ".join(saved)
            self._show_status_bar_msg(f"Program files {saved} saved successfully")
        return True

    @staticmethod
    def _save_program_file(file_path, doc):
        """Saves program file."""
        try:
            with open(file_path, "w") as file:
                file.write(doc.toPlainText())
            return None
        except IOError as err:
            return err

    def init_programfile_list(self):
        """List program files in QTreeView."""
        for name in ("Main program file", "Additional program files"):
            item = QStandardItem(name + "        ")
            item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable & ~Qt.ItemFlag.ItemIsSelectable)
            self.programfiles_model.appendRow(item)
        # Setup 'Main' item
        index = self.programfiles_model.index(0, 0)
        widget = ToolBarWidget("Main program file", self)
        widget.tool_bar.addActions(
            [
                self._ui.actionNew_main_program_file,
                self._ui.actionSelect_main_program_file,
                self._ui.actionRemove_all_program_files,
            ]
        )
        widget.tool_bar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._ui.treeView_programfiles.setIndexWidget(index, widget)
        # Setup 'Additional...' item
        index = self.programfiles_model.index(1, 0)
        widget = ToolBarWidget("Additional program files", self)
        widget.tool_bar.addActions(
            [
                self._ui.actionNew_program_file,
                self._ui.actionAdd_program_file,
                self._ui.actionAdd_program_directory,
                self._ui.actionRemove_selected_program_files,
            ]
        )
        widget.tool_bar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._ui.treeView_programfiles.setIndexWidget(index, widget)
        self._ui.treeView_programfiles.expandAll()
        tool_tip = (
            "<p>Other program files and/or directories (in addition to the main program file) required by the tool.</p>"
            '<p><span style=" font-weight:600;">Tip</span>: '
            "You can Drag &amp; Drop files and/or directories here from your computer.</p>"
        )
        self.programfiles_model.setData(index, tool_tip, role=Qt.ItemDataRole.ToolTipRole)

    def _current_main_program_file(self):
        """Returns the full path to the current main program file or None if there's no main program."""
        index = self.programfiles_model.index(0, 0)
        root_item = self.programfiles_model.itemFromIndex(index)
        if not root_item:
            return None
        if not root_item.hasChildren():
            return None
        return root_item.child(0).data(Qt.ItemDataRole.UserRole)

    def _additional_program_file_list(self):
        return self.spec_dict.get("includes", [])[1:]

    def populate_main_programfile(self, file_path):
        """Adds the main program item into the program files model.

        Args:
            file_path (str): Absolute path to main program file
        """
        index = self.programfiles_model.index(0, 0)
        root_item = self.programfiles_model.itemFromIndex(index)
        root_item.removeRows(0, root_item.rowCount())
        if file_path:  # Do only if the spec has a main program (Executable Tool specs may not have one)
            if self.has_root_directory():
                rel = os.path.relpath(file_path, self.item.root_dir)
                item = QStandardItem(rel)
                full_path = os.path.abspath(os.path.join(self.item.root_dir, file_path))
                item.setData(full_path, Qt.ItemDataRole.UserRole)
            else:
                item = QStandardItem(os.path.basename(file_path))
                item.setData(file_path, Qt.ItemDataRole.UserRole)
            item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)
            item.setData(QFileIconProvider().icon(QFileInfo(file_path)), Qt.ItemDataRole.DecorationRole)
            root_item.appendRow(item)
            tool_tip = f"<p>{self._current_main_program_file()}</p>"
            self.programfiles_model.setData(root_item.child(0).index(), tool_tip, role=Qt.ItemDataRole.ToolTipRole)
            self._ui.treeView_programfiles.selectionModel().setCurrentIndex(
                root_item.child(0).index(), QItemSelectionModel.SelectionFlag.Select
            )
        if self.spec_dict.get("tooltype") == "executable":
            self.optional_widget.set_command_and_shell_edit_disabled_state(bool(file_path))

    def populate_programfile_list(self, names):
        """Adds additional program files into the program files model.

        Args:
            names (list): List of paths to additional program files
        """
        # Find visible indexes, disconnect 'set program file dirty' slots
        visible = set()
        for item in self.programfiles_model.findItems("", Qt.MatchFlag.MatchContains | Qt.MatchFlag.MatchRecursive):
            if not item.rowCount():
                index = self.programfiles_model.indexFromItem(item)
                file_path = self._programfile_path_from_index(index)
                if self._ui.treeView_programfiles.isExpanded(index.parent()):
                    visible.add(file_path)
                doc = self._programfile_documents.get(file_path)
                slot = self._programfile_set_dirty_slots.get(file_path)
                if None not in (doc, slot):
                    doc.modificationChanged.disconnect(slot)
        # Repopulate
        index = self.programfiles_model.index(1, 0)
        root_item = self.programfiles_model.itemFromIndex(index)
        root_item.removeRows(0, root_item.rowCount())
        components = _collect_components(names)
        _build_tree(root_item, components)
        # Re-expand visible, connect 'set program file dirty' slots
        self._programfile_set_dirty_slots.clear()
        for item in self.programfiles_model.findItems("", Qt.MatchFlag.MatchContains | Qt.MatchFlag.MatchRecursive):
            if not item.rowCount():
                index = self.programfiles_model.indexFromItem(item)
                file_path = self._programfile_path_from_index(index)
                if file_path:
                    tool_tip = f"<p>{file_path}</p>"
                    self.programfiles_model.setData(index, tool_tip, role=Qt.ItemDataRole.ToolTipRole)
                if file_path in visible:
                    self._ui.treeView_programfiles.expand(index.parent())
                self._programfile_set_dirty_slots[file_path] = index
                doc = self._programfile_documents.get(file_path)
                if doc is not None:
                    slot = self._programfile_set_dirty_slots[file_path] = (
                        lambda dirty, ind=index: self._set_program_file_dirty(ind, dirty)
                    )
                    doc.modificationChanged.connect(slot)
                    self._set_program_file_dirty(index, doc.isModified())

    def _set_program_file_dirty(self, index, dirty):
        """Appends a * to the file name in program files tree view if dirty.

        Args:
            index (QModelIndex): Index in program files model
            dirty (bool): True or False
        """
        file_path = index.data(Qt.ItemDataRole.UserRole)
        # Main program file items have their full path as UserRole data, additional
        # program file items have only their file name as UserRole data
        if (
            self.has_root_directory()
            and os.path.exists(file_path)
            and os.path.samefile(self._current_main_program_file(), file_path)
        ):
            name = os.path.relpath(file_path, self.item.root_dir)
        else:
            name = os.path.basename(file_path)
        if dirty:
            name += "*"
        self.programfiles_model.setData(index, name, role=Qt.ItemDataRole.DisplayRole)

    def init_io_file_list(self):
        for name in ("Input files", "Optional input files", "Output files"):
            item = QStandardItem(name + "         ")
            item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable & ~Qt.ItemFlag.ItemIsSelectable)
            self.io_files_model.appendRow(item)
        # Setup 'Input' item
        index = self.io_files_model.index(0, 0)
        widget = ToolBarWidget("Input files", self, True)
        widget.tool_bar.addActions([self._ui.actionAdd_input_files, self._ui.actionRemove_selected_input_files])
        widget.tool_bar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._ui.treeView_io_files.setIndexWidget(index, widget)
        # Setup 'Optional' item
        index = self.io_files_model.index(1, 0)
        widget = ToolBarWidget("Optional input files", self, True)
        widget.tool_bar.addActions([self._ui.actionAdd_opt_input_files, self._ui.actionRemove_selected_opt_input_files])
        widget.tool_bar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._ui.treeView_io_files.setIndexWidget(index, widget)
        # Setup 'Output' item
        index = self.io_files_model.index(2, 0)
        widget = ToolBarWidget("Output files", self, True)
        widget.tool_bar.addActions([self._ui.actionAdd_output_files, self._ui.actionRemove_selected_output_files])
        widget.tool_bar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)
        self._ui.treeView_io_files.setIndexWidget(index, widget)

    def _input_file_list(self):
        return self.spec_dict.get("inputfiles", [])

    def _opt_input_file_list(self):
        return self.spec_dict.get("inputfiles_opt", [])

    def _output_file_list(self):
        return self.spec_dict.get("outputfiles", [])

    def _populate_io_files_list(self, index, names):
        root_item = self.io_files_model.itemFromIndex(index)
        root_item.removeRows(0, root_item.rowCount())
        for name in names:
            item = QStandardItem(name)
            item.setData(QFileIconProvider().icon(QFileInfo(name)), Qt.ItemDataRole.DecorationRole)
            root_item.appendRow(item)
        self._ui.treeView_io_files.expand(index)

    def populate_inputfiles_list(self, names):
        """List input files in QTreeView."""
        self.spec_dict["inputfiles"] = names
        index = self.io_files_model.index(0, 0)
        self._populate_io_files_list(index, names)

    def populate_inputfiles_opt_list(self, names):
        """List optional input files in QTreeView."""
        self.spec_dict["inputfiles_opt"] = names
        index = self.io_files_model.index(1, 0)
        self._populate_io_files_list(index, names)

    def populate_outputfiles_list(self, names):
        """List output files in QTreeView."""
        self.spec_dict["outputfiles"] = names
        index = self.io_files_model.index(2, 0)
        self._populate_io_files_list(index, names)

    def connect_signals(self):
        """Connects signals to slots."""
        self._ui.actionNew_main_program_file.triggered.connect(self.new_main_program_file)
        self._ui.actionSelect_main_program_file.triggered.connect(self.browse_main_program_file)
        self._ui.actionRemove_all_program_files.triggered.connect(self.remove_all_program_files)
        self._ui.actionNew_program_file.triggered.connect(self.new_program_file)
        self._ui.actionAdd_program_file.triggered.connect(self.show_add_program_files_dialog)
        self._ui.actionAdd_program_directory.triggered.connect(self.show_add_program_dirs_dialog)
        self._ui.actionRemove_selected_program_files.triggered.connect(self.remove_program_files)
        self._ui.treeView_programfiles.files_dropped.connect(self.add_dropped_program_files)
        self._ui.treeView_programfiles.doubleClicked.connect(self.open_program_file)
        self._ui.actionAdd_input_files.triggered.connect(self.add_inputfiles)
        self._ui.actionRemove_selected_input_files.triggered.connect(self.remove_inputfiles)
        self._ui.actionAdd_opt_input_files.triggered.connect(self.add_inputfiles_opt)
        self._ui.actionRemove_selected_opt_input_files.triggered.connect(self.remove_inputfiles_opt)
        self._ui.actionAdd_output_files.triggered.connect(self.add_outputfiles)
        self._ui.actionRemove_selected_output_files.triggered.connect(self.remove_outputfiles)
        # Enable removing items from QTreeViews by pressing the Delete key
        self._ui.treeView_programfiles.del_key_pressed.connect(self.remove_program_files_with_del)
        self._ui.treeView_io_files.del_key_pressed.connect(self.remove_io_files_with_del)
        # Push undo commands
        self._ui.comboBox_tooltype.currentIndexChanged.connect(self._push_change_tooltype_command)
        self._ui.comboBox_tooltype.currentIndexChanged.connect(self._show_optional_widget)
        self._ui.lineEdit_args.textEdited.connect(self._push_change_args_command)
        self._ui.lineEdit_args.editingFinished.connect(self._finish_changing_args)
        self.io_files_model.dataChanged.connect(self._push_io_file_renamed_command)
        # Selection changed
        self._ui.treeView_programfiles.selectionModel().selectionChanged.connect(
            self._handle_programfile_selection_changed
        )
        self._ui.treeView_io_files.selectionModel().selectionChanged.connect(self._handle_io_file_selection_changed)
        self.programfiles_model.rowsInserted.connect(self._enable_additional_program_files_actions)
        self.programfiles_model.rowsRemoved.connect(self._enable_additional_program_files_actions)

    @Slot(int)
    def _push_change_tooltype_command(self, index):
        new_type = TOOL_TYPES[index].lower()
        old_type = self.spec_dict.get("tooltype", "")
        if new_type == old_type:
            return
        self._undo_stack.push(
            ChangeSpecPropertyCommand(self._set_tooltype, new_type, old_type, "change Tool specification type")
        )

    def _set_tooltype(self, value):
        self.spec_dict["tooltype"] = value
        self._ui.textEdit_program.set_lexer_name(value)
        index = next(iter(k for k, t in enumerate(TOOL_TYPES) if t.lower() == value), -1)
        self._ui.comboBox_tooltype.setCurrentIndex(index)
        self.clear_execution_settings(value)

    def clear_execution_settings(self, spec_type):
        """Updates the execution settings dict based on selected tool spec type.
        Sets the default execution settings into self.spec_dict when a Python,
        Julia, or Executable is selected as type. Removes execution settings from
        self.spec_dict when any other Tool spec type is selected."""
        if "execution_settings" in self.spec_dict:
            self.spec_dict.pop("execution_settings")  # Clear execution settings
        if spec_type == "python":
            self.spec_dict["execution_settings"] = {
                "use_jupyter_console": bool(
                    int(self._toolbox.qsettings().value("appSettings/usePythonKernel", defaultValue="0"))
                ),
                "kernel_spec_name": self._toolbox.qsettings().value("appSettings/pythonKernel", defaultValue=""),
                "env": "",
                "executable": self._toolbox.qsettings().value("appSettings/pythonPath", defaultValue=""),
            }
        elif spec_type == "julia":
            self.spec_dict["execution_settings"] = {
                "use_jupyter_console": bool(
                    int(self._toolbox.qsettings().value("appSettings/useJuliaKernel", defaultValue="0"))
                ),
                "kernel_spec_name": self._toolbox.qsettings().value("appSettings/juliaKernel", defaultValue=""),
                "project": self._toolbox.qsettings().value("appSettings/juliaProjectPath", defaultValue=""),
                "env": "",
                "executable": self._toolbox.qsettings().value("appSettings/juliaPath", defaultValue=""),
            }
        elif spec_type == "executable":
            self.spec_dict["execution_settings"] = {"cmd": "", "shell": ""}

    @Slot(int)
    def push_change_kernel_spec_command(self, index):
        """Changes selected kernel name for Julia and Python Tool Specs."""
        toolspectype = self.spec_dict.get("tooltype", "")
        opt_widget = self._get_optional_widget(toolspectype)
        item = opt_widget.kernel_spec_model.item(index)
        if not item.data():
            new_kernel_spec = ""
        else:
            new_kernel_spec = item.data()["kernel_spec_name"]
        previous_kernel_spec = self.spec_dict["execution_settings"]["kernel_spec_name"]
        if new_kernel_spec == previous_kernel_spec:
            return
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self._set_kernel_spec, new_kernel_spec, previous_kernel_spec, "change kernel spec"
            )
        )

    def _set_kernel_spec(self, value):
        self.spec_dict["execution_settings"]["kernel_spec_name"] = value
        opt_widget = self._get_optional_widget("python")  # TODO: BUG? "python" should be toolspectype?
        row = opt_widget.find_index_by_data(value)  # Find row of item in combobox
        if row == -1:
            logging.error(f"Item '{value}' not found in any combobox item's data")
        opt_widget.ui.comboBox_kernel_specs.setCurrentIndex(row)

    @Slot(bool)
    def push_set_jupyter_console_mode(self, new_value):
        """Changes between Basic or Jupyter Console execution modes for Julia and Python Tool Specs."""
        old_value = self.spec_dict["execution_settings"]["use_jupyter_console"]
        if new_value == old_value:
            return
        self._undo_stack.push(
            ChangeSpecPropertyCommand(self._set_jupyter_console_mode, new_value, old_value, "set Jupyter console mode")
        )

    def _set_jupyter_console_mode(self, value):
        self.spec_dict["execution_settings"]["use_jupyter_console"] = value
        toolspectype = self.spec_dict.get("tooltype", "")
        opt_widget = self._get_optional_widget(toolspectype)
        if value:
            opt_widget.ui.radioButton_jupyter_console.setChecked(True)
        else:
            opt_widget.ui.radioButton_basic_console.setChecked(True)
        opt_widget.set_ui_for_jupyter_console(value)

    @Slot(str)
    def push_change_executable(self, executable):
        """Changes Julia or Python executable path for Julia and Python Tool Specs.

        Args:
            executable (str): path to executable
        """
        old_value = self.spec_dict["execution_settings"]["executable"]
        new_value = executable.strip()
        if new_value == old_value:
            return
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self._set_executable, new_value, old_value, "change path of executable", CommandId.EXECUTABLE_UPDATE
            )
        )

    @Slot()
    def finish_changing_executable(self):
        """Seals latest undo stack command."""
        self._undo_stack.push(SealCommand(CommandId.EXECUTABLE_UPDATE.value))

    def _set_executable(self, value):
        self.spec_dict["execution_settings"]["executable"] = value
        toolspectype = self.spec_dict.get("tooltype", "")
        opt_widget = self._get_optional_widget(toolspectype)
        opt_widget.set_executable(value)

    @Slot()
    def push_change_project(self):
        """Changes Julia project path for Julia Tool Specs."""
        old_value = self.spec_dict["execution_settings"]["project"]
        toolspectype = self.spec_dict.get("tooltype", "")
        opt_widget = self._get_optional_widget(toolspectype)
        new_value = opt_widget.get_julia_project()
        if new_value == old_value:
            return
        self._undo_stack.push(
            ChangeSpecPropertyCommand(self._set_julia_project, new_value, old_value, "change Julia project")
        )

    def _set_julia_project(self, value):
        self.spec_dict["execution_settings"]["project"] = value
        toolspectype = self.spec_dict.get("tooltype", "")
        opt_widget = self._get_optional_widget(toolspectype)
        opt_widget.set_julia_project(value)

    @Slot(int)
    def push_change_shell_command(self, index):
        """Changes shell for Executable Tool Specs."""
        toolspectype = self.spec_dict.get("tooltype", "")
        opt_widget = self._get_optional_widget(toolspectype)
        new_shell = opt_widget.shells[index] if index != 0 else ""
        old_shell = self.spec_dict["execution_settings"]["shell"]
        if new_shell == old_shell:
            return
        self._undo_stack.push(ChangeSpecPropertyCommand(self._set_shell, new_shell, old_shell, "change shell"))

    def _set_shell(self, value):
        self.spec_dict["execution_settings"]["shell"] = value
        opt_widget = self._get_optional_widget("executable")
        index = next(iter(k for k, t in enumerate(opt_widget.shells) if t.lower() == value), 0)
        opt_widget.ui.comboBox_shell.setCurrentIndex(index)

    @Slot(str)
    def _push_change_args_command(self, args):
        """Changes cmd line args for all Tool Specs.

        Args:
            args (str): new args
        """
        old_value = self.spec_dict.get("cmdline_args", [])
        new_value = split_cmdline_args(args)
        if new_value == old_value:
            return
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self._set_cmdline_args,
                new_value,
                old_value,
                "change command line args",
                command_id=CommandId.ARGS_UPDATE,
            )
        )

    @Slot()
    def _finish_changing_args(self):
        """Seals latest undo stack command."""
        self._undo_stack.push(SealCommand(CommandId.ARGS_UPDATE.value))

    def _set_cmdline_args(self, value):
        if split_cmdline_args(self._ui.lineEdit_args.text()) == value:
            return
        self.spec_dict["cmdline_args"] = value
        self._ui.lineEdit_args.setText(" ".join(value))

    @Slot(str)
    def push_change_executable_command(self, command):
        """Changes command for Executable Tool Specs.

        Args:
            command (str): new command
        """
        old_value = self.spec_dict["execution_settings"]["cmd"]
        new_value = command.strip()
        if new_value == old_value:
            return
        self._undo_stack.push(
            ChangeSpecPropertyCommand(self._set_cmd, new_value, old_value, "change command", CommandId.COMMAND_UPDATE)
        )

    @Slot()
    def finish_changing_command(self):
        """Seals latest undo stack command."""
        self._undo_stack.push(SealCommand(CommandId.COMMAND_UPDATE.value))

    def _set_cmd(self, value):
        self.spec_dict["execution_settings"]["cmd"] = value
        opt_widget = self._get_optional_widget("executable")
        opt_widget.ui.lineEdit_command.setText(value)

    def _push_change_main_program_file_command(self, new_value):
        """Pushes a command that changes the main program file to undo stack.

        Args:
            new_value (str, optional): absolute path to the main program file; if None, the file will be removed
        """
        old_program_files = self.spec_dict.get("includes", [])
        if self.includes_main_path is not None:
            old_program_files = [os.path.join(self.includes_main_path, f) for f in old_program_files]
        old_value = old_program_files[0] if old_program_files else ""
        if new_value is not None:
            if same_path(old_value, new_value):
                return
            new_program_files = old_program_files.copy()
            if new_program_files:
                new_program_files[0] = new_value
            elif new_value:
                new_program_files.append(new_value)
        else:
            new_program_files = []
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self._set_program_files, new_program_files, old_program_files, "change main program file"
            )
        )

    def _set_program_files(self, program_files):
        """Sets program files.

        Args:
            program_files (list of str): List of paths. First item is the absolute path to main program file,
                the remaining items are path stubs relative to the item root directory or to includes main path.
        """
        main_program_file = program_files[0] if program_files else ""
        if self.has_root_directory():
            additional_program_files = [os.path.relpath(file, self.includes_main_path) for file in program_files[1:]]
            main_program_path = os.path.relpath(main_program_file, self.item.root_dir) if main_program_file else ""
        else:
            self.includes_main_path = os.path.dirname(next(iter(f for f in program_files if f), ""))
            additional_program_files = [os.path.relpath(file, self.includes_main_path) for file in program_files[1:]]
            main_program_path = os.path.basename(main_program_file)
        self._set_main_program_file(main_program_file)
        self.populate_programfile_list(additional_program_files)
        self.spec_dict["includes"] = [main_program_path, *additional_program_files]

    def _set_main_program_file(self, file_path):
        """Sets main program file and dumps its contents into the text edit.

        Args:
            file_path (str): Absolute path
        """
        self.populate_main_programfile(file_path)
        self._label_main_path.setText(self.includes_main_path)

    @Slot(QItemSelection, QItemSelection)
    def _handle_programfile_selection_changed(self, _selected, _deselected):
        # Set remove action enabled
        indexes = self._selected_program_file_indexes()
        self._ui.actionRemove_selected_program_files.setEnabled(bool(indexes))
        # Load selected file code on text edit
        current = self._ui.treeView_programfiles.selectionModel().currentIndex()
        if not current.parent().isValid() or self.programfiles_model.rowCount(current) != 0:
            self._clear_program_text_edit()
            return
        self._load_programfile_in_editor(current)

    def _programfile_path_from_index(self, index):
        """Returns absolute path to a file pointed by index.

        Args:
            index (QModelIndex): index

        Returns:
            str: file path
        """
        components = _path_components_from_index(index)
        if not components or not self.includes_main_path:
            return ""
        return os.path.join(self.includes_main_path, *components)

    def _clear_program_text_edit(self):
        """Clears contents of the text editor."""
        empty_doc = QTextDocument(self._ui.textEdit_program)
        self._ui.textEdit_program.setDocument(empty_doc)
        self._ui.textEdit_program.setEnabled(False)
        self._ui.textEdit_program.file_selected(False)
        self._ui.dockWidget_program.setWindowTitle("No file selected")

    def _load_programfile_in_editor(self, index):
        """Loads contents of a program file into the text editor.

        Args:
            index (QModelIndex): index in programfiles_model
        """
        file_path = self._programfile_path_from_index(index)
        if not os.path.isfile(file_path):
            self._show_status_bar_msg(f"Program file {file_path} is not valid")
            self._clear_program_text_edit()
            return
        if file_path not in self._programfile_documents:
            try:
                with open(file_path, "r") as file:
                    text = file.read()
            except (IOError, UnicodeDecodeError) as e:
                self._show_status_bar_msg(str(e))
                return
            document = self._programfile_documents[file_path] = QTextDocument(self)
            document.setPlainText(text)
            document.setModified(False)
            slot = self._programfile_set_dirty_slots[file_path] = lambda dirty, ind=index: self._set_program_file_dirty(
                ind, dirty
            )
            document.modificationChanged.connect(slot)
            document.modificationChanged.connect(self._update_window_modified)
        else:
            document = self._programfile_documents[file_path]
        self._ui.textEdit_program.setDocument(document)
        self._ui.textEdit_program.setEnabled(True)
        self._ui.textEdit_program.file_selected(True)
        self._ui.dockWidget_program.setWindowTitle(os.path.basename(file_path) + "[*]")

    def _start_dir(self):
        """Returns path to main program directory if available, or to project directory otherwise."""
        if self._current_main_program_file() is not None:
            current_main_program_dir, _ = os.path.split(self._current_main_program_file())
        else:
            current_main_program_dir = None
        return self._project.project_dir if not current_main_program_dir else current_main_program_dir

    @Slot(bool)
    def browse_main_program_file(self, _=False):
        """Opens a file dialog where user can select the path of the main program file."""
        filter = self._get_filetype_filter()
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        answer = QFileDialog.getOpenFileName(self, "Select existing main program file", self._start_dir(), filter)
        file_path = answer[0]
        existing_file_paths = [
            os.path.join(self.includes_main_path, i)
            for i in self.spec_dict.get("includes", [])
            if os.path.exists(os.path.join(self.includes_main_path, i))
        ]
        if not file_path:  # Cancel button clicked
            return
        for i, existing_file in enumerate(existing_file_paths):
            if not existing_file:
                continue
            if os.path.samefile(file_path, existing_file):
                if i != 0:
                    msg = QMessageBox(self)
                    msg.setIcon(QMessageBox.Icon.Information)
                    msg.setWindowTitle(self.windowTitle())
                    msg.setText("Can't choose an existing additional\nprogram file as the main program file")
                    msg.setStandardButtons(QMessageBox.StandardButton.Ok)
                    msg.exec()
                return
        self._push_change_main_program_file_command(file_path)

    @Slot(bool)
    def new_main_program_file(self, _=False):
        """Creates a new blank main program file."""
        filter = self._get_filetype_filter()
        # noinspection PyCallByClass
        answer = QFileDialog.getSaveFileName(self, "Create new main program file", self._start_dir(), filter)
        file_path = answer[0]
        existing_file_paths = [os.path.join(self.includes_main_path, i) for i in self.spec_dict.get("includes", [])]
        if not file_path:  # Cancel button clicked
            return
        # Remove file if it exists. getSaveFileName has asked confirmation for us.
        try:
            os.remove(file_path)
        except OSError:
            pass
        try:
            with open(file_path, "w"):
                pass
        except OSError:
            msg = "Please check directory permissions."
            # noinspection PyTypeChecker, PyArgumentList, PyCallByClass
            QMessageBox.information(self, "Creating file failed", msg)
            return
        for i in existing_file_paths:
            if not i:
                continue
            if os.path.samefile(file_path, i):
                msg = "File with same file path already\nin use as an additional program file."
                QMessageBox.information(self, "Creating file failed", msg)
                return
        self._push_change_main_program_file_command(file_path)

    @Slot()
    def new_program_file(self):
        """Let user create a new program file for this tool specification."""
        path = self.includes_main_path if self.includes_main_path else self._project.project_dir
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        dir_path = QFileDialog.getSaveFileName(self, "Create program file", path, "*.*")
        file_path = dir_path[0]
        existing_file_paths = [os.path.join(self.includes_main_path, i) for i in self.spec_dict.get("includes", [])]
        if file_path == "":  # Cancel button clicked
            return
        # create file. NOTE: getSaveFileName does the 'check for existence' for us
        open(file_path, "w").close()
        for i in existing_file_paths:
            if os.path.samefile(file_path, i):
                msg = "File with same file path already in use."
                QMessageBox.information(self, "Creating file failed", msg)
                return
        self.add_program_files(file_path)

    @Slot(bool)
    def show_add_program_files_dialog(self, _=False):
        """Let user select program files for this tool specification."""
        path = self.includes_main_path if self.includes_main_path else self._project.project_dir
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        answer = QFileDialog.getOpenFileNames(self, "Add program file", path, "*.*")
        file_paths = answer[0]
        if not file_paths:  # Cancel button clicked
            return
        for file_path in file_paths:
            if os.path.samefile(self._current_main_program_file(), file_path):
                file_paths.remove(file_path)
                msg = QMessageBox(self)
                msg.setWindowTitle(self.windowTitle())
                msg.setIcon(QMessageBox.Icon.Information)
                msg.setStandardButtons(QMessageBox.StandardButton.Ok)
                if len(file_paths) == 0:
                    msg.setText("Can't add main program file\nto additional program files")
                    msg.exec()
                    return
                msg.setText("Skipped main program file")
                msg.exec()
        self.add_program_files(*file_paths)

    @Slot(bool)
    def show_add_program_dirs_dialog(self, _=False):
        """Let user select a program directory for this tool specification.
        All files and sub-directories will be added to the program files.
        """
        path = self.includes_main_path if self.includes_main_path else self._project.project_dir
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        answer = QFileDialog.getExistingDirectory(self, "Select a directory to add to program files", path)
        file_paths = []
        main_program_file = self._current_main_program_file()
        for root, _, files in os.walk(answer):
            for file in files:
                os.path.abspath(os.path.join(root, file))
                if not os.path.samefile(main_program_file, os.path.abspath(os.path.join(root, file))):
                    file_paths.append(os.path.abspath(os.path.join(root, file)))
        self.add_program_files(*file_paths)

    @Slot(list)
    def add_dropped_program_files(self, file_paths):
        """Adds dropped file paths to Source files list."""
        self.add_program_files(*file_paths)

    def _get_filetype_filter(self):
        """Returns a filter for the QFileDialog based on the selected tool spec type."""
        tooltype = self.spec_dict.get("tooltype", "")
        if tooltype == "python":
            return "Python (*.py);;*.*"
        elif tooltype == "julia":
            return "Julia (*.jl);;*.*"
        elif tooltype == "gams":
            return "Gams (*.gms);;*.*"
        else:
            return "*.*"

    def _validate_additional_program_files(self, new_files, old_program_files):
        valid_files = []
        dupes = []
        invalid = []
        for file in new_files:
            if file in old_program_files:
                dupes.append(os.path.basename(file))
                continue
            if self.includes_main_path is not None:
                common_prefix = os.path.commonprefix([os.path.abspath(self.includes_main_path), os.path.abspath(file)])
                if not same_path(common_prefix, self.includes_main_path):
                    invalid.append(os.path.basename(file))
                    continue
            valid_files.append(file)
        if dupes:
            dupes = ", ".join(dupes)
            self._show_status_bar_msg(f"Program file(s) '{dupes}' already added")
        if invalid:
            invalid = ", ".join(invalid)
            self._show_status_bar_msg(f"Program file(s) '{invalid}' not in main directory")
        return valid_files

    def add_program_files(self, *new_files):
        """Appends program files.

        Args:
            *new_files (str): Absolute paths to append.
        """
        old_program_files = self.spec_dict.get("includes", [])
        if self.includes_main_path is not None:
            old_program_files = [os.path.join(self.includes_main_path, f) for f in old_program_files]
        new_files = self._validate_additional_program_files(list(new_files), old_program_files)
        if not new_files:
            return
        new_program_files = old_program_files.copy() + new_files
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self._set_program_files, new_program_files, old_program_files, "add program files"
            )
        )

    @Slot()
    def remove_program_files_with_del(self):
        """Adds support for deleting items with the Delete key."""
        self.remove_program_files()

    def _selected_program_file_indexes(self):
        """Returns selected program file indexes in a dict excluding the main program file"""
        indexes = {index for index in self._ui.treeView_programfiles.selectedIndexes() if index.parent().isValid()}
        # discard main program file
        parent = self.programfiles_model.index(0, 0)
        indexes.discard(self.programfiles_model.index(0, 0, parent))
        # keep only leaves
        return {ind for ind in indexes if self.programfiles_model.rowCount(ind) == 0}

    @Slot(bool)
    def remove_all_program_files(self, _=False):
        """Removes all program files including main program. Useful with Executable Tool Specs."""
        self._ui.treeView_programfiles.selectAll()
        inds = set(self._ui.treeView_programfiles.selectedIndexes())
        indexes = {ind for ind in inds if not self.programfiles_model.rowCount(ind)}  # keep only leaves
        if not indexes:
            self._show_status_bar_msg("No program files to remove")
            return
        old_program_files = self.spec_dict.get("includes", [])
        new_program_files = []
        if self.includes_main_path is not None:
            old_program_files = [os.path.join(self.includes_main_path, f) for f in old_program_files]
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self._set_program_files, new_program_files, old_program_files, "remove all program files"
            )
        )

    @Slot(bool)
    def remove_program_files(self, _=False):
        """Removes selected program files from program_file list."""
        indexes = self._selected_program_file_indexes()
        if not indexes:  # Nothing removable selected
            self._show_status_bar_msg("Please select program files to remove")
            return
        removed_files = {os.path.join(*_path_components_from_index(ind)) for ind in indexes}
        old_program_files = self.spec_dict.get("includes", [])
        old_program_files = [f.replace(os.sep, "/") for f in old_program_files]
        removed_files = [f.replace(os.sep, "/") for f in removed_files]
        new_program_files = [f for f in old_program_files if f not in removed_files]
        if self.includes_main_path is not None:
            old_program_files = [os.path.join(self.includes_main_path, f) for f in old_program_files]
            new_program_files = [os.path.join(self.includes_main_path, f) for f in new_program_files]
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self._set_program_files, new_program_files, old_program_files, "remove program files"
            )
        )

    @busy_effect
    @Slot(QModelIndex)
    def open_program_file(self, index):
        """Opens a program file in default program."""
        if not index.isValid() or self.programfiles_model.rowCount(index) > 0:
            return
        program_file = self._programfile_path_from_index(index)
        _, ext = os.path.splitext(program_file)
        if ext in [".bat", ".exe"]:
            self._toolbox.msg_warning.emit(
                f"Sorry, opening files with extension <b>{ext}</b> not implemented. Please open the file manually."
            )
            return
        url = "file:///" + os.path.join(self.includes_main_path, program_file)
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        res = open_url(url)
        if not res:
            self._toolbox.msg_error.emit(f"Failed to open file: <b>{program_file}</b>")

    @Slot(bool)
    def _enable_additional_program_files_actions(self, _=False):
        """Enables actions depending on whether Main program file exists."""
        main_program_file_exists = self._current_main_program_file() is not None
        self._ui.actionNew_program_file.setEnabled(main_program_file_exists)
        self._ui.actionAdd_program_file.setEnabled(main_program_file_exists)
        self._ui.actionAdd_program_directory.setEnabled(main_program_file_exists)

    @Slot(bool)
    def add_inputfiles(self, _=False):
        """Opens a file dialog for selecting input files for this tool specification."""
        msg = (
            "Add an input file or a directory required by your program. Wildcards "
            "<b>are not</b> supported.<br/><br/>"
            "Examples:<br/>"
            "<b>data.csv</b> -> File is copied to the same work directory as the main program.<br/>"
            "<b>input/data.csv</b> -> Creates subdirectory /input to work directory and "
            "copies file data.csv there.<br/>"
            "<b>output/</b> -> Creates an empty directory into the work directory.<br/><br/>"
        )
        title = "Add input item"
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        answer = QInputDialog.getText(
            self, title, msg, flags=Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint
        )
        file_name = answer[0]
        if not file_name:  # Cancel button clicked
            return
        old_files = self.spec_dict.get("inputfiles", [])
        new_files = old_files + [file_name]
        self._undo_stack.push(
            ChangeSpecPropertyCommand(self.populate_inputfiles_list, new_files, old_files, "add input file")
        )

    @Slot(bool)
    def add_inputfiles_opt(self, _=False):
        """Opens a file dialog for selecting optional input files for this tool specification."""
        msg = (
            "Add optional input files that may be utilized by your program. <br/>"
            "Wildcards are supported.<br/><br/>"
            "Examples:<br/>"
            "<b>data.csv</b> -> If found, file is copied to the same work directory as the main program.<br/>"
            "<b>*.csv</b> -> All found CSV files are copied to the same work directory as the main program.<br/>"
            "<b>input/data_?.dat</b> -> All found files matching the pattern 'data_?.dat' will be copied to <br/>"
            "input/ subdirectory under the same work directory as the main program.<br/><br/>"
        )
        title = "Add optional input item"
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        answer = QInputDialog.getText(
            self, title, msg, flags=Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint
        )
        file_name = answer[0]
        if not file_name:  # Cancel button clicked
            return
        old_files = self.spec_dict.get("inputfiles_opt", [])
        new_files = old_files + [file_name]
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self.populate_inputfiles_opt_list, new_files, old_files, "add optional input file"
            )
        )

    @Slot(bool)
    def add_outputfiles(self, _=False):
        """Opens a file dialog for selecting output files for this tool specification."""
        msg = (
            "Add output files that will be archived into the Tool results directory after the <br/>"
            "Tool specification has finished execution. Wildcards are supported.<br/><br/>"
            "Examples:<br/>"
            "<b>results.csv</b> -> File is copied from work directory into results.<br/> "
            "<b>*.csv</b> -> All CSV files will copied into results.<br/> "
            "<b>output/*.gdx</b> -> All GDX files from the work subdirectory /output will be copied into <br/>"
            "results /output subdirectory.<br/><br/>"
        )
        title = "Add output item"
        # noinspection PyCallByClass, PyTypeChecker, PyArgumentList
        answer = QInputDialog.getText(
            self, title, msg, flags=Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint
        )
        file_name = answer[0]
        if not file_name:  # Cancel button clicked
            return
        old_files = self.spec_dict.get("outputfiles", [])
        new_files = old_files + [file_name]
        self._undo_stack.push(
            ChangeSpecPropertyCommand(self.populate_outputfiles_list, new_files, old_files, "add output file")
        )

    @Slot(QModelIndex, QModelIndex, list)
    def _push_io_file_renamed_command(self, top_left, bottom_right, roles):
        """Pushes a command to rename input/output files to undo stack.

        Args:
            top_left (QModelIndex): top left index to renamed files
            bottom_right (QModelIndex): bottom right index to renamed files
            roles (list of int):
        """
        if Qt.ItemDataRole.DisplayRole not in roles:
            return
        parent_index = top_left.parent()
        if parent_index.row() == 0:
            old_files = self.spec_dict.get("inputfiles", [])
            callback = self.populate_inputfiles_list
            message = "rename input file"
        elif parent_index.row() == 1:
            old_files = self.spec_dict.get("inputfiles_opt", [])
            callback = self.populate_inputfiles_opt_list
            message = "rename optional input file"
        else:
            old_files = self.spec_dict.get("outputfiles", [])
            callback = self.populate_outputfiles_list
            message = "rename output file"
        new_files = list(old_files)
        for j in range(top_left.column(), bottom_right.column() + 1):
            for i in range(top_left.row(), bottom_right.row() + 1):
                new_files[i] = self.io_files_model.index(i, j, parent_index).data()
        self._undo_stack.push(ChangeSpecPropertyCommand(callback, new_files, old_files, message))

    def _selected_io_file_indexes(self, parent):
        indexes = self._ui.treeView_io_files.selectedIndexes()
        return [ind for ind in indexes if ind.parent() == parent]

    @Slot(QItemSelection, QItemSelection)
    def _handle_io_file_selection_changed(self, _deselected, _selected):
        parent = self.io_files_model.index(0, 0)
        indexes = self._selected_io_file_indexes(parent)
        self._ui.actionRemove_selected_input_files.setEnabled(bool(indexes))
        parent = self.io_files_model.index(1, 0)
        indexes = self._selected_io_file_indexes(parent)
        self._ui.actionRemove_selected_opt_input_files.setEnabled(bool(indexes))
        parent = self.io_files_model.index(2, 0)
        indexes = self._selected_io_file_indexes(parent)
        self._ui.actionRemove_selected_output_files.setEnabled(bool(indexes))

    @Slot(bool)
    def remove_inputfiles(self, _=False):
        """Removes selected input files from list. Does not remove anything if no items selected."""
        parent = self.io_files_model.index(0, 0)
        indexes = self._selected_io_file_indexes(parent)
        if not indexes:  # Nothing selected
            self._show_status_bar_msg("Please select the input files to remove")
            return
        removed_files = {ind.data(Qt.ItemDataRole.DisplayRole) for ind in indexes}
        old_files = self.spec_dict.get("inputfiles", [])
        new_files = [f for f in old_files if f not in removed_files]
        self._undo_stack.push(
            ChangeSpecPropertyCommand(self.populate_inputfiles_list, new_files, old_files, "remove input files")
        )

    @Slot(bool)
    def remove_inputfiles_opt(self, _=False):
        """Removes selected optional input files from list. Does not remove anything if no items selected."""
        parent = self.io_files_model.index(1, 0)
        indexes = self._selected_io_file_indexes(parent)
        if not indexes:  # Nothing selected
            self._show_status_bar_msg("Please select the optional input files to remove")
            return
        removed_files = {ind.data(Qt.ItemDataRole.DisplayRole) for ind in indexes}
        old_files = self.spec_dict.get("inputfiles_opt", [])
        new_files = [f for f in old_files if f not in removed_files]
        self._undo_stack.push(
            ChangeSpecPropertyCommand(
                self.populate_inputfiles_opt_list, new_files, old_files, "remove optional input files"
            )
        )

    @Slot(bool)
    def remove_outputfiles(self, _=False):
        """Removes selected output files from list. Does not remove anything if no items selected."""
        parent = self.io_files_model.index(2, 0)
        indexes = self._selected_io_file_indexes(parent)
        if not indexes:  # Nothing selected
            self._show_status_bar_msg("Please select the output files to remove")
            return
        removed_files = {ind.data(Qt.ItemDataRole.DisplayRole) for ind in indexes}
        old_files = self.spec_dict.get("outputfiles", [])
        new_files = [f for f in old_files if f not in removed_files]
        self._undo_stack.push(
            ChangeSpecPropertyCommand(self.populate_outputfiles_list, new_files, old_files, "remove output files")
        )

    @Slot()
    def remove_io_files_with_del(self):
        """Support for deleting items with the Delete key."""
        self.remove_inputfiles()
        self.remove_inputfiles_opt()
        self.remove_outputfiles()


def _path_components(path):
    """
    Args:
        path(str): a filesystem path, e.g. (/home/user/spine/toolbox/file.py)

    Yields:
        str: path components in 'reverse' order, e.g., file.py, toolbox, spine, user, home
    """
    while path:
        path, component = os.path.split(path)
        yield component


def _collect_components(names):
    """
    Args:
        names (list(str)): list of filepaths

    Returns:
        dict: mapping folders to contents recursively
    """
    components = {}
    for name in names:
        d = components
        for component in reversed(list(_path_components(name))):
            d = d.setdefault(component, {})
    return components


def _build_tree(root, components):
    """Appends rows from given components to given root item.

    Args:
        root (QStandardItem): root item
        components (dict): mapping folders to contents recursively
    """
    nodes = []
    leafs = []
    for parent, children in components.items():
        item = QStandardItem(parent)
        item.setData(parent, Qt.ItemDataRole.UserRole)
        item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)
        _build_tree(item, children)
        if item.hasChildren():
            nodes.append(item)
            item.setData(QFileIconProvider().icon(QFileIconProvider.IconType.Folder), Qt.ItemDataRole.DecorationRole)
        else:
            item.setData(QFileIconProvider().icon(QFileInfo(parent)), Qt.ItemDataRole.DecorationRole)
            leafs.append(item)
    for item in sorted(nodes, key=methodcaller("text")):
        root.appendRow(item)
    for item in sorted(leafs, key=methodcaller("text")):
        root.appendRow(item)


def _path_components_from_index(index):
    components = []
    while index.parent().isValid():
        components.insert(0, index.data(Qt.ItemDataRole.UserRole))
        index = index.parent()
    return components



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Tool specification classes."""
from collections import OrderedDict
import copy
import os.path
from spine_engine.project_item.project_item_specification import ProjectItemSpecification
from spine_engine.utils.command_line_arguments import split_cmdline_args
from .item_info import ItemInfo
from .tool_instance import ExecutableToolInstance, GAMSToolInstance, JuliaToolInstance, PythonToolInstance

# Tool types
TOOL_TYPES = ["Julia", "Python", "GAMS", "Executable"]

# Required and optional keywords for Tool specification dictionaries
REQUIRED_KEYS = ["name", "tooltype", "includes"]
OPTIONAL_KEYS = [
    "description",
    "short_name",
    "inputfiles",
    "inputfiles_opt",
    "outputfiles",
    "cmdline_args",
    "execution_settings",
]
LIST_REQUIRED_KEYS = ["includes", "inputfiles", "inputfiles_opt", "outputfiles"]  # These should be lists


def make_specification(definition, app_settings, logger):
    """Deserializes and constructs a tool specification from definition.

    Args:
        definition (dict): a dictionary containing the serialized specification.
        app_settings (QSettings): Toolbox settings
        logger (LoggerInterface): a logger

    Returns:
        ToolSpecification: a tool specification constructed from the given definition,
            or None if there was an error
    """
    if not definition["includes_main_path"]:
        path = None
    else:
        definition["includes_main_path"] = path = definition.setdefault("includes_main_path", ".").replace(
            "/", os.path.sep
        )
        if not os.path.isabs(path):
            definition_file_path = definition["definition_file_path"]
            path = os.path.normpath(os.path.join(os.path.dirname(definition_file_path), path))
        definition["includes"] = [src_file.replace("/", os.path.sep) for src_file in definition["includes"]]
    try:
        _tooltype = definition["tooltype"].lower()
    except KeyError:
        logger.msg_error.emit(
            "No tool type defined in tool definition file. Supported types "
            "are 'python', 'gams', 'julia' and 'executable'"
        )
        return None
    if _tooltype == "julia":
        spec = JuliaTool.load(path, definition, app_settings, logger)
    elif _tooltype == "python":
        spec = PythonTool.load(path, definition, app_settings, logger)
    elif _tooltype == "gams":
        spec = GAMSTool.load(path, definition, app_settings, logger)
    elif _tooltype == "executable":
        spec = ExecutableTool.load(path, definition, app_settings, logger)
    else:
        logger.msg_warning.emit(f"Tool type <b>{_tooltype}</b> not available")
        return None
    return spec


class ToolSpecification(ProjectItemSpecification):
    """Super class for all tool specifications."""

    def __init__(
        self,
        name,
        tooltype,
        path,
        includes,
        settings,
        logger,
        description=None,
        inputfiles=None,
        inputfiles_opt=None,
        outputfiles=None,
        cmdline_args=None,
    ):
        """

        Args:
            name (str): Tool specification name
            tooltype (str): Type of Tool Specification (e.g. Python, Julia, ..)
            path (str): Path to Tool specification
            includes (list): List of files belonging to the tool specification (relative to 'path')
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): a logger instance
            description (str): Description of the Tool specification
            inputfiles (list): List of required data files
            inputfiles_opt (list, optional): List of optional data files (wildcards may be used)
            outputfiles (list, optional): List of output files (wildcards may be used)
            cmdline_args (list, optional): Tool Specification command line arguments
        """
        super().__init__(name, description, item_type=ItemInfo.item_type())
        self._settings = settings
        self._logger = logger
        self.tooltype = tooltype
        self.path = path
        self.includes = includes
        if cmdline_args is not None:
            if isinstance(cmdline_args, str):
                # Old tool spec files may have the command line arguments as plain strings.
                self.cmdline_args = split_cmdline_args(cmdline_args)
            else:
                self.cmdline_args = cmdline_args
        else:
            self.cmdline_args = []
        self.inputfiles = set(inputfiles) if inputfiles else set()
        self.inputfiles_opt = set(inputfiles_opt) if inputfiles_opt else set()
        self.outputfiles = set(outputfiles) if outputfiles else set()
        self.return_codes = {}

    def _includes_main_path_relative(self):
        return os.path.relpath(self.path, os.path.dirname(self.definition_file_path)).replace(os.path.sep, "/")

    def clone(self):
        spec_dict = copy.deepcopy(self.to_dict())
        spec_dict["definition_file_path"] = self.definition_file_path
        return make_specification(spec_dict, self._settings, self._logger)

    def to_dict(self):
        return {
            "name": self.name,
            "tooltype": self.tooltype,
            "includes": [src_file.replace(os.path.sep, "/") for src_file in self.includes],
            "description": self.description,
            "inputfiles": sorted(self.inputfiles),
            "inputfiles_opt": sorted(self.inputfiles_opt),
            "outputfiles": sorted(self.outputfiles),
            "cmdline_args": self.cmdline_args,
            "includes_main_path": self._includes_main_path_relative(),
        }

    @staticmethod
    def _definition_local_entries():
        """See base class"""
        return [("execution_settings",)]

    def init_execution_settings(self):
        """Updates Tool specifications by adding the default execution settings dict for this specification."""

    def is_equivalent(self, other):
        """See base class."""
        for k, v in other.__dict__.items():
            if k in LIST_REQUIRED_KEYS:
                if set(self.__dict__[k]) != set(v):
                    return False
            else:
                if self.__dict__[k] != v:
                    return False
        return True

    def set_return_code(self, code, description):
        """Sets a return code and an associated text description for the tool specification.

        Args:
            code (int): Return code
            description (str): Description
        """
        self.return_codes[code] = description

    @staticmethod
    def check_definition(data, logger):
        """Checks that a tool specification contains
        the required keys and that it is in correct format.

        Args:
            data (dict): Tool specification
            logger (LoggerInterface): A logger instance

        Returns:
            dict: definition or None if there was a problem in the tool definition.
        """
        kwargs = {}
        for p in REQUIRED_KEYS + OPTIONAL_KEYS:
            try:
                kwargs[p] = data[p]
            except KeyError:
                if p in REQUIRED_KEYS:
                    logger.msg_error.emit(f"Required keyword '{p}' missing")
                    return None
            # Check that some values are lists
            if p in LIST_REQUIRED_KEYS:
                try:
                    if not isinstance(data[p], list):
                        logger.msg_error.emit(f"Keyword '{p}' value must be a list")
                        return None
                except KeyError:
                    pass
        return kwargs

    def create_tool_instance(self, basedir, kill_completed_processes, logger, owner):
        """Returns an instance of this tool specification that is configured to run in the given directory.

        Args:
            basedir (str): Path to the directory where the instance will run
            kill_completed_processes (bool): whether to kill completed persistent processes
            logger (LoggerInterface): Logger
            owner (ExecutableItemBase): Project item that owns the instance
        """
        raise NotImplementedError()


class GAMSTool(ToolSpecification):
    """Class for GAMS tool specifications."""

    def __init__(
        self,
        name,
        tooltype,
        path,
        includes,
        settings,
        logger,
        description=None,
        inputfiles=None,
        inputfiles_opt=None,
        outputfiles=None,
        cmdline_args=None,
    ):
        """

        Args:
            name (str): GAMS Tool name
            tooltype (str): Tool specification type
            path (str): Path to model main file
            includes (list): List of files belonging to the tool (relative to 'path').
                First file in the list is the main GAMS program.
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): a logger instance
            description (str): GAMS Tool description
            inputfiles (list): List of required data files
            inputfiles_opt (list, optional): List of optional data files (wildcards may be used)
            outputfiles (list, optional): List of output files (wildcards may be used)
            cmdline_args (list, optional): GAMS Tool Specification command line arguments
        """
        super().__init__(
            name,
            tooltype,
            path,
            includes,
            settings,
            logger,
            description,
            inputfiles,
            inputfiles_opt,
            outputfiles,
            cmdline_args,
        )
        main_file_path = includes[0]
        # Add .lst file to list of output files
        subdir, main_file = os.path.split(main_file_path)
        self.lst_file = os.path.splitext(main_file)[0] + ".lst"
        self.outputfiles.add(self.lst_file)
        self.main_prgm = main_file_path
        self.gams_options = OrderedDict()
        self.return_codes = {
            0: "Normal return",
            1: "Solver is to be called the system should never return this number",  # ??
            2: "There was a compilation error",
            3: "There was an execution error",
            4: "System limits were reached",
            5: "There was a file error",
            6: "There was a parameter error",
            7: "There was a licensing error",
            8: "There was a GAMS system error",
            9: "GAMS could not be started",
            10: "Out of memory",
            11: "Out of disk",
            62097: "Simulation interrupted by user",  # Not official
        }

    def make_debug_project(self, basedir):
        """Make a GAMS project file and return an anchor for the execution item log.

        Args:
            basedir (str): Full path to work or source dir depending on execution mode

        Returns:
            Anchor to gamside project anchor or None if operation failed
        """
        prj_file_path = os.path.abspath(os.path.join(self.path, self.short_name + "_autocreated.gpr"))
        try:
            self.make_gamside_project_file(prj_file_path, basedir)
        except OSError:
            return None
        anchor = (
            "<a style='color:#99CCFF;' title='"
            + prj_file_path
            + "' href='file:///"
            + prj_file_path
            + "'>Click here to debug in GAMSIDE</a>"
        )
        return anchor

    def make_gamside_project_file(self, prj_file_path, basedir):
        """Make a GAMSIDE project file for debugging.

        Args:
            prj_file_path (str): Full path to GAMSIDE project file
            basedir (str): Full path to work or source dir depending on execution mode

        Returns:
            Full path to GAMSIDE project file or None if operation failed
        """
        lst_file_path = os.path.join(basedir, self.lst_file)  # We need the latest from work dir or from source dir
        main_prgm_path = os.path.join(self.path, self.main_prgm)  # Always get the one from source dir
        with open(prj_file_path, "w") as f:
            f.write("[PROJECT]\n\n")
            f.write("[MRUFILES]\n")  # Most Recently Used (MRU)
            f.write("1=" + lst_file_path + "\n\n")
            f.write("[OPENWINDOW_1]\n")
            f.write("FILE0=" + main_prgm_path + "\n")
            f.write("FILE1=" + lst_file_path + "\n")
            f.write("FILE2=" + main_prgm_path + "\n")
            f.write("MAXIM=0\n")
            f.write("TOP=0\n")
            f.write("LEFT=0\n")
            f.write("HEIGHT=600\n")
            f.write("WIDTH=1000\n")

    @staticmethod
    def load(path, data, settings, logger):
        """Creates a GAMSTool according to a tool specification file.

        Args:
            path (str): Base path to tool files
            data (dict): Dictionary of tool definitions
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): A logger instance

        Returns:
            GAMSTool instance or None if there was a problem in the tool specification file.
        """
        kwargs = GAMSTool.check_definition(data, logger)
        if kwargs is not None:
            # Return an executable model instance
            return GAMSTool(path=path, settings=settings, logger=logger, **kwargs)
        return None

    def create_tool_instance(self, basedir, kill_completed_processes, logger, owner):
        """See base class."""
        return GAMSToolInstance(self, basedir, self._settings, logger, owner)


class JuliaTool(ToolSpecification):
    """Class for Julia tool specifications."""

    def __init__(
        self,
        name,
        tooltype,
        path,
        includes,
        settings,
        logger,
        description=None,
        inputfiles=None,
        inputfiles_opt=None,
        outputfiles=None,
        cmdline_args=None,
        execution_settings=None,
    ):
        """
        Args:
            name (str): Julia Tool name
            tooltype (str): Tool specification type
            path (str): Path to model main file
            includes (list): List of files belonging to the tool (relative to 'path')
                First file in the list is the main Julia program.
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): A logger instance
            description (str): Julia Tool description
            inputfiles (list): List of required data files
            inputfiles_opt (list, optional): List of optional data files (wildcards may be used)
            outputfiles (list, optional): List of output files (wildcards may be used)
            cmdline_args (list, optional): Julia Tool Specification command line arguments
            execution_settings (dict, optional): Julia tool spec execution settings
        """
        super().__init__(
            name,
            tooltype,
            path,
            includes,
            settings,
            logger,
            description,
            inputfiles,
            inputfiles_opt,
            outputfiles,
            cmdline_args,
        )
        self.main_prgm = includes[0]
        self.julia_options = OrderedDict()
        self.execution_settings = execution_settings if execution_settings is not None else {}
        self.return_codes = {0: "Normal return", -1: "Failure"}

    def to_dict(self):
        """Adds execution settings dict to Tool spec dict."""
        d = super().to_dict()
        d["execution_settings"] = self.execution_settings
        return d

    def set_execution_settings(self, use_julia_jupyter_console, julia_exe, julia_project, julia_kernel):
        """Sets execution settings.

        Args:
            use_julia_jupyter_console (bool): True for Jupyter Console, False for Basic Console
            julia_exe (str): Abs. path to Julia executable
            julia_project (str): Julia project
            julia_kernel (str): Julia kernel for Jupyter Console
        """
        d = {}
        d["kernel_spec_name"] = julia_kernel
        d["env"] = ""
        d["use_jupyter_console"] = use_julia_jupyter_console
        d["executable"] = julia_exe
        d["project"] = julia_project
        self.execution_settings = d

    def init_execution_settings(self):
        """Initializes execution_setting instance attribute to default settings."""
        if not self.execution_settings:
            # Use global (default) execution settings from Settings->Tools
            # This part is for providing support for Julia Tool specs that do not have
            # the execution_settings dict yet
            d = {}
            d["kernel_spec_name"] = self._settings.value("appSettings/juliaKernel", defaultValue="")
            d["env"] = ""
            d["use_jupyter_console"] = bool(
                int(self._settings.value("appSettings/useJuliaKernel", defaultValue="0"))
            )  # bool(int(str))
            d["executable"] = self._settings.value("appSettings/juliaPath", defaultValue="")
            d["project"] = self._settings.value("appSettings/juliaProjectPath", defaultValue="")
            self.execution_settings = d

    @staticmethod
    def load(path, data, settings, logger):
        """Creates a JuliaTool according to a tool specification file.

        Args:
            path (str): Base path to tool files
            data (dict): Dictionary of tool definitions
            settings (QSetting): Toolbox settings
            logger (LoggerInterface): A logger instance

        Returns:
            JuliaTool instance or None if there was a problem in the tool definition file.
        """
        kwargs = JuliaTool.check_definition(data, logger)
        if kwargs is not None:
            # Return an executable model instance
            return JuliaTool(path=path, settings=settings, logger=logger, **kwargs)
        return None

    def create_tool_instance(self, basedir, kill_completed_processes, logger, owner):
        """See base class."""
        return JuliaToolInstance(self, basedir, self._settings, kill_completed_processes, logger, owner)


class PythonTool(ToolSpecification):
    """Class for Python tool specifications."""

    def __init__(
        self,
        name,
        tooltype,
        path,
        includes,
        settings,
        logger,
        description=None,
        inputfiles=None,
        inputfiles_opt=None,
        outputfiles=None,
        cmdline_args=None,
        execution_settings=None,
    ):
        """
        Args:

            name (str): Python Tool name
            tooltype (str): Tool specification type
            path (str): Path to model main file
            includes (list): List of files belonging to the tool (relative to 'path').
                First file in the list is the main Python program.
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): A logger instance
            description (str): Python Tool description
            inputfiles (list): List of required data files
            inputfiles_opt (list, optional): List of optional data files (wildcards may be used)
            outputfiles (list, optional): List of output files (wildcards may be used)
            cmdline_args (list, optional): Python Tool Specification command line arguments
            execution_settings (dict, optional): Python tool spec execution settings
        """
        super().__init__(
            name,
            tooltype,
            path,
            includes,
            settings,
            logger,
            description,
            inputfiles,
            inputfiles_opt,
            outputfiles,
            cmdline_args,
        )
        self.main_prgm = includes[0]
        self.python_options = OrderedDict()
        self.execution_settings = execution_settings if execution_settings is not None else {}
        self.return_codes = {0: "Normal return", -1: "Failure"}  # Not official

    def to_dict(self):
        """Adds kernel spec settings dict to Tool spec dict."""
        d = super().to_dict()
        d["execution_settings"] = self.execution_settings
        return d

    def set_execution_settings(self, use_python_jupyter_console, python_exe, python_kernel, env):
        """Sets execution settings.

        Args:
            use_python_jupyter_console (bool): True for Jupyter Console, False for Basic Console
            python_exe (str): Abs. path to Python executable
            python_kernel (str): Julia kernel for Jupyter Console
            env (str): empty string for regular kernels, 'conda' for Conda kernels
        """
        d = {}
        d["kernel_spec_name"] = python_kernel
        d["env"] = env
        d["use_jupyter_console"] = use_python_jupyter_console
        d["executable"] = python_exe
        self.execution_settings = d

    def init_execution_settings(self):
        """Sets the execution_setting instance attribute to defaults if this
        Python Tool spec has not been updated yet.

        Returns:
            void
        """
        if not self.execution_settings:
            # Use global (default) execution settings from Settings->Tools
            # This part is for providing support for Python Tool specs that do not have
            # the execution_settings dict yet
            d = {}
            d["kernel_spec_name"] = self._settings.value("appSettings/pythonKernel", defaultValue="")
            d["env"] = ""
            d["use_jupyter_console"] = bool(
                int(self._settings.value("appSettings/usePythonKernel", defaultValue="0"))
            )  # bool(int(str))
            d["executable"] = self._settings.value("appSettings/pythonPath", defaultValue="")
            self.execution_settings = d

    @staticmethod
    def load(path, data, settings, logger):
        """Creates a PythonTool according to a tool specification file.

        Args:
            path (str): Base path to tool files
            data (dict): Dictionary of tool definitions
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): A logger instance

        Returns:
            PythonTool instance or None if there was a problem in the tool definition file.
        """
        kwargs = PythonTool.check_definition(data, logger)
        if kwargs is not None:
            # Return an executable model instance
            return PythonTool(path=path, settings=settings, logger=logger, **kwargs)
        return None

    def create_tool_instance(self, basedir, kill_completed_processes, logger, owner):
        """See base class."""
        return PythonToolInstance(self, basedir, self._settings, kill_completed_processes, logger, owner)


class ExecutableTool(ToolSpecification):
    """Class for Executable tool specifications."""

    def __init__(
        self,
        name,
        tooltype,
        path,
        includes,
        settings,
        logger,
        description=None,
        inputfiles=None,
        inputfiles_opt=None,
        outputfiles=None,
        cmdline_args=None,
        execution_settings=None,
        definition_file_path=None,
    ):
        """
        Args:

            name (str): Tool name
            tooltype (str): Tool specification type
            path (str): Path to main script file
            includes (list): List of files belonging to the tool (relative to 'path')
                First file in the list is the main program file.
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): A logger instance
            description (str): Tool description
            inputfiles (list): List of required data files
            inputfiles_opt (list, optional): List of optional data files (wildcards may be used)
            outputfiles (list, optional): List of output files (wildcards may be used)
            cmdline_args (list, optional): Executable Tool Specification command line arguments
            execution_settings (dict): Settings for executing a (shell) command instead of a file
            definition_file_path (str): Absolute path to spec definition file. Only used when running a (shell) command
                in 'source' execution mode
        """
        super().__init__(
            name,
            tooltype,
            path,
            includes,
            settings,
            logger,
            description,
            inputfiles,
            inputfiles_opt,
            outputfiles,
            cmdline_args,
        )
        try:
            self.main_prgm = includes[0]
        except IndexError:
            self.main_prgm = None
        self.execution_settings = execution_settings
        if definition_file_path and not path:
            # Set the default execution dir in source execution mode when running a command (no program files)
            # This part is processed when Tool Specs are loaded at app startup or when Spine Engine loads them
            # Note: When a Tool Spec is saved in Tool Spec Editor, definition_file_path == None
            # Default execution dir is the directory of the definition file path
            self.default_execution_dir, _ = os.path.split(definition_file_path)
        self.options = OrderedDict()
        self.return_codes = {0: "Normal exit", 1: "Error happened"}

    def to_dict(self):
        """Adds execution settings dict to Executable Tool spec dicts."""
        d = super().to_dict()
        d["execution_settings"] = self.execution_settings
        return d

    def _includes_main_path_relative(self):
        if not self.path:
            return None
        return super()._includes_main_path_relative()

    def init_execution_settings(self):
        """Updates old Executable Tool specifications by adding the
        default execution settings dict for this specification.

        Returns:
            void
        """
        if not self.execution_settings:
            d = {}
            d["cmd"] = ""
            d["shell"] = ""
            self.execution_settings = d

    @staticmethod
    def load(path, data, settings, logger):
        """Creates an ExecutableTool according to a tool specification file.

        Args:
            path (str): Base path to tool files
            data (dict): Tool specification
            settings (QSettings): Toolbox settings
            logger (LoggerInterface): A logger instance

        Returns:
            ExecutableTool instance or None if there was a problem in the tool specification.
        """
        kwargs = ExecutableTool.check_definition(data, logger)
        if kwargs is not None:
            # Add definition_file_path. 'data' does not have definition_file_path
            # when creating a new Tool spec or editing an existing one in Tool Spec Editor.
            kwargs["definition_file_path"] = data.get("definition_file_path", None)
            # Return an executable model instance
            return ExecutableTool(path=path, settings=settings, logger=logger, **kwargs)
        return None

    def create_tool_instance(self, basedir, kill_completed_processes, logger, owner):
        """See base class."""
        return ExecutableToolInstance(self, basedir, self._settings, logger, owner)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Classes for custom context menus and pop-up menus."""
import os
from PySide6.QtCore import QUrl, Slot
from spinetoolbox.helpers import open_url
from spinetoolbox.widgets.custom_menus import ItemSpecificationMenu


class ToolSpecificationMenu(ItemSpecificationMenu):
    """Menu class for Tool specifications."""

    def __init__(self, toolbox, index, item=None):
        """
        Args:
            toolbox (ToolboxUI): the toolbox that request this menu
            index (QModelIndex): the index from specification model
            item (Tool, optional): The item that owns the menu
        """
        super().__init__(toolbox, index, item)
        self.addSeparator()
        self.add_action("Open main program file", self._open_main_program_file)
        self.add_action("Open main program directory", self._open_main_program_dir)

    @Slot()
    def _open_main_program_file(self):
        spec = self.parent().specification_model.specification(self.index.row())
        if not spec.path or not os.path.isdir(spec.path):
            if not spec.includes:
                self._toolbox.msg_warning.emit("No main program to open")
                return
            self._toolbox.msg_error.emit(
                f"Opening Tool spec main program file <b>{spec.includes[0]}</b> failed. "
                f"Main program directory does not exist."
            )
            return
        file_path = os.path.join(spec.path, spec.includes[0])
        # Check that file exists
        if not os.path.isfile(file_path):
            self._toolbox.msg_error.emit(f"Tool spec main program file <b>{file_path}</b> not found.")
            return
        ext = os.path.splitext(os.path.split(file_path)[1])[1]
        if ext in [".bat", ".exe"]:
            self._toolbox.msg_warning.emit(
                f"Sorry, opening files with extension <b>{ext}</b> not supported. " f"Please open the file manually."
            )
            return
        main_program_url = "file:///" + file_path
        res = open_url(main_program_url)
        if not res:
            filename, file_extension = os.path.splitext(file_path)
            self._toolbox.msg_error.emit(
                f"Unable to open Tool specification main program file {filename}. "
                f"Make sure that <b>{file_extension}</b> "
                f"files are associated with an editor. E.g. on Windows "
                f"10, go to Control Panel -> Default Programs to do this."
            )

    @Slot()
    def _open_main_program_dir(self):
        tool_specification = self._toolbox.specification_model.specification(self.index.row())
        if not tool_specification.includes:
            self._toolbox.msg_warning.emit("No main program directory to open")
            return
        if not tool_specification.path:
            self._toolbox.msg_error.emit("Main program directory does not exist. Fix this in Tool spec editor.")
            return
        path_url = "file:///" + tool_specification.path
        self._toolbox.open_anchor(QUrl(path_url, QUrl.TolerantMode))



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains utilities to scan for Tool's output resources."""
from spine_engine.project_item.project_item_resource import (
    file_resource,
    file_resource_in_pack,
    transient_file_resource,
)
from .utils import find_last_output_files, is_pattern


def scan_for_resources(provider, tool_specification, output_dir):
    """
    Returns a list of resources, i.e. the output files produced by the tool.

    For each pattern or path in the tool specification's output files list, we try and find matches
    in the results directory.

    If no match, we advertise a resource with an empty url, where the label contains the non-found
    pattern or path.

    Args:
        provider (ProjectItem or ExecutableItem): resource provider item
        tool_specification (ToolSpecification): spec
        output_dir (str): path to tool's output folder

    Returns:
        list: a list of Tool's output resources
    """
    resources = []
    if tool_specification is None:
        return []
    last_output_files = find_last_output_files(tool_specification.outputfiles, output_dir)
    for out_file_label in tool_specification.outputfiles:
        latest_files = last_output_files.get(out_file_label, [])
        make_resource = file_resource if not is_pattern(out_file_label) else file_resource_in_pack
        for out_file in latest_files:
            resources.append(make_resource(provider.name, file_path=out_file, label=out_file_label))
        if not latest_files:
            make_resource = transient_file_resource if not is_pattern(out_file_label) else file_resource_in_pack
            resources.append(make_resource(provider.name, label=out_file_label))
    return resources



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for Data transformer's :class:`DataTransformerSpecification`."""
import unittest
from spine_items.data_transformer.data_transformer_specification import DataTransformerSpecification
from spine_items.data_transformer.settings import EntityClassRenamingSettings


class TestDataTransformerSpecification(unittest.TestCase):
    def test_is_equivalent(self):
        dts1 = DataTransformerSpecification("DT SPEC")
        dts2 = DataTransformerSpecification("DT SPEC")
        self.assertTrue(dts1.is_equivalent(dts2))

    def test_to_dict(self):
        dts = DataTransformerSpecification("DT SPEC")
        dts_dict = dts.to_dict()
        self.assertEqual(4, len(dts_dict.keys()))
        self.assertEqual("DT SPEC", dts_dict["name"])
        self.assertEqual("Data Transformer", dts_dict["item_type"])
        self.assertIsNone(dts_dict["description"])
        self.assertIsNone(dts_dict["filter"])

    def test_from_dict(self):
        dts_dict = DataTransformerSpecification("DT SPEC").to_dict()
        dts = DataTransformerSpecification.from_dict(dts_dict, None)
        self.assertTrue(isinstance(dts, DataTransformerSpecification))
        # Make one with Settings
        rename_map = {"a": "b"}
        settings = EntityClassRenamingSettings(rename_map)
        dts2 = DataTransformerSpecification("DT WITH SETTINGS", settings=settings)
        dts2_dict = dts2.to_dict()
        dts2_reloaded = DataTransformerSpecification.from_dict(dts2_dict, None)
        self.assertTrue(isinstance(dts2_reloaded, DataTransformerSpecification))


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Tool's executable item and support functionality."""
from contextlib import ExitStack
import datetime
import fnmatch
import glob
import os
import os.path
import pathlib
import re
import shutil
import time
import uuid
from spine_engine.config import TOOL_OUTPUT_DIR
from spine_engine.project_item.project_item_resource import (
    expand_cmd_line_args,
    labelled_resource_args,
    make_cmd_line_arg,
)
from spine_engine.spine_engine import ItemExecutionFinishState
from spine_engine.utils.helpers import (
    create_log_file_timestamp,
    resolve_gams_executable,
    resolve_julia_executable,
    write_filter_id_file,
)
from spine_engine.utils.serialization import deserialize_path
from ..db_writer_executable_item_base import DBWriterExecutableItemBase
from ..utils import generate_filter_subdirectory_name
from .item_info import ItemInfo
from .output_resources import scan_for_resources
from .utils import file_paths_from_resources, find_file, flatten_file_path_duplicates, is_pattern, make_dir_if_necessary

_ANSI_ESCAPE = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


class ExecutableItem(DBWriterExecutableItemBase):
    """Tool project item's executable parts."""

    _MAX_RETRIES = 3

    def __init__(
        self,
        name,
        work_dir,
        tool_specification,
        cmd_line_args,
        options,
        kill_completed_processes,
        log_process_output,
        group_id,
        project_dir,
        logger,
    ):
        """
        Args:
            name (str): item's name
            work_dir (str): an absolute path to Spine Toolbox work directory
                or None if the Tool should not execute in work directory
            tool_specification (ToolSpecification): a tool specification
            cmd_line_args (list): a list of command line argument to pass to the tool instance
            options (dict): misc tool options. See ``Tool`` for details.
            kill_completed_processes (bool): whether to kill completed persistent processes after execution
            log_process_output (bool): whether to log process output to a file
            group_id (str or None): execution group identifier
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, project_dir, logger, group_id=group_id)
        self._work_dir = work_dir
        output_dir = deserialize_path(options.get("output_directory"), project_dir)
        self._output_dir = str(pathlib.Path(self._data_dir, TOOL_OUTPUT_DIR)) if not output_dir else output_dir
        self._tool_specification = tool_specification
        self._cmd_line_args = cmd_line_args
        self._options = options
        self._kill_completed_processes = kill_completed_processes
        self._tool_instance = None
        self._log_file_paths = {}
        self.max_log_file_size = 128 * 1024
        if log_process_output:
            self._logger.msg_standard_execution.connect(self._log_execution_msg)
            self._logger.msg_persistent_execution.connect(self._log_execution_msg)
            self._logger.msg_kernel_execution.connect(self._log_execution_msg)

    def _make_log_file_path(self, filter_id, timestamp, part_nb):
        log_file_path = os.path.abspath(
            os.path.join(self._logs_dir, "_".join([timestamp, "execution", str(part_nb)]) + ".log")
        )
        with open(log_file_path, "w") as log_file:
            header = [
                "### Spine execution log file",
                f"### Item name: {self._name}",
                f"### Filter id: {filter_id}",
                f"### Part: {part_nb}",
            ]
            log_file.write("\n".join(header) + "\n\n")
        return log_file_path

    def _log_execution_msg(self, msg):
        if msg["type"] not in ("stdin", "stdout", "stderr"):
            return
        filter_id = msg["filter_id"]
        if filter_id not in self._log_file_paths:
            timestamp = create_log_file_timestamp()
            part_nb = 1
            log_file_path = self._make_log_file_path(filter_id, timestamp, part_nb)
            self._log_file_paths[filter_id] = (log_file_path, timestamp, part_nb)
        log_file_path, timestamp, part_nb = self._log_file_paths[filter_id]
        if os.path.isfile(log_file_path) and os.path.getsize(log_file_path) > self.max_log_file_size:
            part_nb += 1
            log_file_path = self._make_log_file_path(filter_id, timestamp, part_nb)
            self._log_file_paths[filter_id] = (log_file_path, timestamp, part_nb)
        with open(log_file_path, "a", encoding="utf-8") as log_file:
            # TODO: use win_unicode_console instead of encoding="utf-8" above
            data = _filter_ansi_escape(msg["data"])
            for line in data.split("\n"):
                if line:
                    log_file.write(line + "\n")

    @property
    def options(self):
        return self._options

    @property
    def cmd_line_args(self):
        return self._cmd_line_args

    @staticmethod
    def item_type():
        """Returns the item's type identifier string."""
        return ItemInfo.item_type()

    def stop_execution(self):
        """Stops executing this Tool."""
        super().stop_execution()
        if self._tool_instance is not None and self._tool_instance.is_running():
            self._tool_instance.terminate_instance()
            self._tool_instance = None

    def _copy_input_files(self, paths, execution_dir):
        """Copies input files from given paths to work or source
        directory, depending on where the Tool specification
        requires them to be.

        Args:
            paths (dict): key is path to destination file, value is path to source file
            execution_dir (str): absolute path to the execution directory

        Returns:
            bool: True if the operation was successful, False otherwise
        """
        n_copied_files = 0
        # Make sure that execution_dir (work dir) exists if there's something to copy.
        # Needed by Executable tools running a shell command
        if not make_dir_if_necessary(paths, execution_dir):
            self._logger.msg_error.emit(f"Creating directory <b>{execution_dir}</b> failed")
            return False
        for dst, src_path in paths.items():
            file_anchor = (
                f"<a style='color:#BB99FF;' title='{src_path}' href='file:///{src_path}'>"
                + f"{os.path.basename(src_path)}</a>"
            )
            if not os.path.exists(src_path):
                self._logger.msg_error.emit(f"\tFile <b>{file_anchor}</b> does not exist")
                return False
            # Join work directory path to dst (dst is the filename including possible subfolders, e.g. 'input/f.csv')
            dst_path = os.path.abspath(os.path.join(execution_dir, dst))
            # Create subdirectories if necessary
            dst_subdir, _ = os.path.split(dst)
            if dst_subdir:
                # Create subdirectory structure to work or source directory
                work_subdir_path = os.path.abspath(os.path.join(execution_dir, dst_subdir))
                if not os.path.exists(work_subdir_path):
                    try:
                        os.makedirs(work_subdir_path, exist_ok=True)
                    except OSError:
                        self._logger.msg_error.emit(f"[OSError] Creating directory <b>{work_subdir_path}</b> failed.")
                        return False
            try:
                shutil.copyfile(src_path, dst_path)
                n_copied_files += 1
            except shutil.SameFileError:
                # Happens in source dir exec mode when DC is passing a file as reference from the main program dir to
                # source dir, which are equal
                self._logger.msg_warning.emit("\tNo need to copy. File already available.")
            except OSError as e:
                self._logger.msg_error.emit(f"Copying file <b>{file_anchor}</b> to <b>{dst_path}</b> failed")
                self._logger.msg_error.emit(f"{e}")
                if e.errno == 22:
                    msg = (
                        "The reason might be:\n"
                        "[1] The destination file already exists and it cannot be "
                        "overwritten because it is locked by Julia or some other application.\n"
                        "[2] You don't have the necessary permissions to overwrite the file.\n"
                        "To solve the problem, you can try the following:\n[1] Execute the Tool in work "
                        "directory.\n[2] If you are executing a Julia Tool with Julia 0.6.x, upgrade to "
                        "Julia 0.7 or newer.\n"
                        "[3] Close any other background application(s) that may have locked the file.\n"
                        "And try again.\n"
                    )
                    self._logger.msg_warning.emit(msg)
                return False
        self._logger.msg.emit(f"\tCopied <b>{n_copied_files}</b> input file(s)")
        return True

    def _copy_optional_input_files(self, paths, execution_dir):
        """Copies optional input files from given paths to work or source
        directory, depending on where the Tool specification requires
        them to be.

        Args:
            paths (dict): key is the source path, value is the destination path
            execution_dir (str): Absolute path to work or source directory

        Returns:
            bool: False if creating the execution directory failed, True otherwise
        """
        n_copied_files = 0
        # Make sure that execution_dir (work dir) exists if there's something to copy.
        # Needed by Executable tools running a shell command
        if not make_dir_if_necessary(paths, execution_dir):
            self._logger.msg_error.emit(f"Creating directory <b>{execution_dir}</b> failed")
            return False
        for src_path, dst_path in paths.items():
            try:
                shutil.copyfile(src_path, dst_path)
                n_copied_files += 1
            except shutil.SameFileError:
                # Happens in source dir exec mode when DC is passing a file as reference from the main program dir to
                # source dir, which are equal
                self._logger.msg_warning.emit("\tNo need to copy. File already available.")
            except OSError as e:
                self._logger.msg_error.emit(f"Copying optional file <b>{src_path}</b> to <b>{dst_path}</b> failed")
                self._logger.msg_error.emit(f"{e}")
                if e.errno == 22:
                    msg = (
                        "The reason might be:\n"
                        "[1] The destination file already exists and it cannot be "
                        "overwritten because it is locked by Julia or some other application.\n"
                        "[2] You don't have the necessary permissions to overwrite the file.\n"
                        "To solve the problem, you can try the following:\n[1] Execute the Tool in work "
                        "directory.\n[2] If you are executing a Julia Tool with Julia 0.6.x, upgrade to "
                        "Julia 0.7 or newer.\n"
                        "[3] Close any other background application(s) that may have locked the file.\n"
                        "And try again.\n"
                    )
                    self._logger.msg_warning.emit(msg)
        self._logger.msg.emit(f"\tCopied <b>{n_copied_files}</b> optional input file(s)")
        return True

    def _copy_output_files(self, target_dir, execution_dir):
        """Copies Tool specification output files from work directory to given target directory.

        Args:
            target_dir (str): Destination directory for Tool specification output files
            execution_dir (str): path to the execution directory

        Returns:
            tuple: Contains two lists. The first list contains paths to successfully
            copied files. The second list contains paths (or patterns) of Tool specification
            output files that were not found.

        Raises:
            OSError: If creating a directory fails.
        """
        failed_files = []
        saved_files = []
        for pattern in self._tool_specification.outputfiles:
            # Create subdirectories if necessary
            dst_subdir, fname_pattern = os.path.split(pattern)
            target = os.path.abspath(os.path.join(target_dir, dst_subdir))
            if not os.path.exists(target):
                try:
                    os.makedirs(target, exist_ok=True)
                except OSError:
                    self._logger.msg_error.emit(f"[OSError] Creating directory <b>{target}</b> failed.")
                    continue
                self._logger.msg.emit(f"\tCreated result subdirectory <b>{os.path.sep}{dst_subdir}</b>")
            # Check for wildcards in pattern
            if is_pattern(pattern):
                for fname_path in glob.glob(os.path.abspath(os.path.join(execution_dir, pattern))):
                    # fname_path is a full path
                    fname = os.path.split(fname_path)[1]  # File name (no path)
                    dst = os.path.abspath(os.path.join(target, fname))
                    full_fname = os.path.join(dst_subdir, fname)
                    try:
                        shutil.copyfile(fname_path, dst)
                        saved_files.append((full_fname, dst))
                    except OSError:
                        self._logger.msg_error.emit(f"[OSError] Copying pattern {fname_path} to {dst} failed")
                        failed_files.append(full_fname)
            else:
                output_file = os.path.abspath(os.path.join(execution_dir, pattern))
                if not os.path.isfile(output_file):
                    failed_files.append(pattern)
                    continue
                dst = os.path.abspath(os.path.join(target, fname_pattern))
                try:
                    shutil.copyfile(output_file, dst)
                    saved_files.append((pattern, dst))
                except OSError:
                    self._logger.msg_error.emit(f"[OSError] Copying output file {output_file} to {dst} failed")
                    failed_files.append(pattern)
        return saved_files, failed_files

    def _copy_program_files(self, execution_dir):
        """Copies Tool specification source files to base directory.

        Args:
            execution_dir (str): path to execution directory

        Returns:
            bool: True if operation was successful, False otherwise
        """
        n_copied_files = 0
        for source_pattern in self._tool_specification.includes:
            dir_name, file_pattern = os.path.split(source_pattern)
            src_dir = os.path.join(self._tool_specification.path, dir_name)
            dst_dir = os.path.join(execution_dir, dir_name)
            # Create the destination directory
            try:
                os.makedirs(dst_dir, exist_ok=True)
            except OSError:
                self._logger.msg_error.emit(f"Creating directory <b>{dst_dir}</b> failed")
                return False
            # Copy file if necessary
            if file_pattern:
                for src_file in glob.glob(os.path.abspath(os.path.join(src_dir, file_pattern))):
                    dst_file = os.path.abspath(os.path.join(dst_dir, os.path.basename(src_file)))
                    try:
                        shutil.copyfile(src_file, dst_file)
                        n_copied_files += 1
                    except OSError:
                        self._logger.msg_error.emit(f"\tCopying file <b>{src_file}</b> to <b>{dst_file}</b> failed")
                        return False
        if n_copied_files == 0:
            self._logger.msg_warning.emit("Warning: No files copied")
        else:
            self._logger.msg.emit(f"\tCopied <b>{n_copied_files}</b> file(s)")
        return True

    def _create_input_dirs(self, execution_dir):
        """Iterates items in required input files and check
        if there are any directories to create. Create found
        directories directly to work or source directory.

        Args:
            execution_dir (str): the execution directory

        Returns:
            bool: True if the operation was successful, False otherwiseBoolean variable depending on success
        """
        for required_path in self._tool_specification.inputfiles:
            path, filename = os.path.split(required_path)
            if filename:
                continue
            path_to_create = os.path.join(execution_dir, path)
            try:
                os.makedirs(path_to_create, exist_ok=True)
            except OSError:
                self._logger.msg_error.emit(f"[OSError] Creating directory {path_to_create} failed. Check permissions.")
                return False
            self._logger.msg.emit(f"\tDirectory <b>{os.path.sep}{path}</b> created")
        return True

    def _create_output_dirs(self, execution_dir):
        """Makes sure that work directory has the necessary output directories for Tool output files.
        Checks only "outputfiles" list. Alternatively you can add directories to "inputfiles" list
        in the tool definition file.

        Args:
            execution_dir (str): a path to the execution directory

        Returns:
            bool: True for success, False otherwise.
        """
        for out_file_path in self._tool_specification.outputfiles:
            dirname = os.path.split(out_file_path)[0]
            if not dirname:
                continue
            dst_dir = os.path.join(execution_dir, dirname)
            try:
                os.makedirs(dst_dir, exist_ok=True)
            except OSError:
                self._logger.msg_error.emit(f"Creating work output directory '{dst_dir}' failed")
                return False
        return True

    def ready_to_execute(self, settings):
        """See base class.

        Returns False when
        1. Tool has no specification
        2. Python or Julia Kernel spec not selected (jupyter kernel mode)
        3. Julia executable not set and not found in PATH (subprocess mode)

        Returns True otherwise.
        """
        if self._tool_specification is None:
            self._logger.msg_warning.emit(f"Tool <b>{self.name}</b> not ready for execution. No specification.")
            return False
        if self._tool_specification.tooltype.lower() == "python":
            use_jupyter_console = self._tool_specification.execution_settings.get("use_jupyter_console", False)
            kernel_name = self._tool_specification.execution_settings.get("kernel_spec_name", "")
            if use_jupyter_console and not kernel_name:
                self._logger.msg_error.emit("Python kernel missing. Please select it in Tool Specification Editor.")
                return False
            # Note: no check for python path == "" because this should never happen
        elif self._tool_specification.tooltype.lower() == "julia":
            use_jupyter_console = self._tool_specification.execution_settings.get("use_jupyter_console", False)
            kernel_name = self._tool_specification.execution_settings.get("kernel_spec_name", "")
            julia_path = self._tool_specification.execution_settings.get("executable", "")
            if not julia_path:
                julia_path = resolve_julia_executable(settings)
            if use_jupyter_console and not kernel_name:
                self._logger.msg_error.emit("Julia kernel missing. Please select it in Tool Specification Editor.")
                return False
            if not use_jupyter_console and not julia_path:
                self._logger.msg_error.emit(
                    "Julia executable path missing. Please set it in Tool Specification Editor."
                )
                return False
        elif self._tool_specification.tooltype.lower() == "gams":
            gams_path = resolve_gams_executable(settings.value("appSettings/gamsPath", defaultValue=""))
            if not gams_path:
                self._logger.msg_error.emit("Gams not found in PATH. Please set a path to Gams in Settings->Tools.")
                return False
        return True

    def execute(self, forward_resources, backward_resources, lock):
        """See base class.

        Before launching the tool script in a separate instance,
        prepares the execution environment by creating all necessary directories
        and copying input files where needed.
        After execution archives the output files.
        """
        if not super().execute(forward_resources, backward_resources, lock):
            return ItemExecutionFinishState.FAILURE
        execution_dir = _execution_directory(self._work_dir, self._tool_specification)
        if execution_dir is None:
            return ItemExecutionFinishState.FAILURE
        work_or_source = "work" if self._work_dir is not None else "source"
        # Make work/source directory anchor with path as tooltip
        anchor = (
            f"<a style='color:#99CCFF;' title='{execution_dir}'"
            f"href='file:///{execution_dir}'>{work_or_source} directory</a>"
        )
        self._logger.msg.emit(
            f"*** Executing Tool specification <b>{self._tool_specification.name}</b> in {anchor} ***"
        )
        if work_or_source == "work":
            self._logger.msg.emit("*** Copying program files ***")
            if not self._copy_program_files(execution_dir):
                self._logger.msg_error.emit("Copying program files failed")
                return ItemExecutionFinishState.FAILURE
        # Find required input files for ToolInstance (if any)
        if self._tool_specification.inputfiles:
            self._logger.msg.emit("*** Checking Tool specification requirements ***")
            n_dirs, n_files = _count_files_and_dirs(self._tool_specification.inputfiles)
            if n_files > 0:
                self._logger.msg.emit("*** Searching for required input files ***")
                file_paths = flatten_file_path_duplicates(
                    self._find_input_files(forward_resources), self._logger, log_duplicates=True
                )
                not_found = [k for k, v in file_paths.items() if v is None]
                if not_found:
                    self._logger.msg_error.emit(f"Required file(s) <b>{', '.join(not_found)}</b> not found")
                    return ItemExecutionFinishState.FAILURE
                self._logger.msg.emit(f"*** Copying input files to {work_or_source} directory ***")
                # Copy input files to ToolInstance work or source directory
                if not self._copy_input_files(file_paths, execution_dir):
                    self._logger.msg_error.emit("Copying input files failed. Tool execution aborted.")
                    return ItemExecutionFinishState.FAILURE
            if n_dirs > 0:
                self._logger.msg.emit(f"*** Creating input subdirectories to {work_or_source} directory ***")
                if not self._create_input_dirs(execution_dir):
                    # Creating directories failed -> abort
                    self._logger.msg_error.emit("Creating input subdirectories failed. Tool execution aborted.")
                    return ItemExecutionFinishState.FAILURE
        if self._tool_specification.inputfiles_opt:
            self._logger.msg.emit("*** Searching for optional input files ***")
            optional_file_paths = self._find_optional_input_files(forward_resources)
            for k, v in optional_file_paths.items():
                self._logger.msg.emit(f"\tFound <b>{len(v)}</b> files matching pattern <b>{k}</b>")
            optional_file_copy_paths = self._optional_output_destination_paths(optional_file_paths, execution_dir)
            if not self._copy_optional_input_files(optional_file_copy_paths, execution_dir):
                return False
        if not self._create_output_dirs(execution_dir):
            self._logger.msg_error.emit("Creating output subdirectories failed. Tool execution aborted.")
            return ItemExecutionFinishState.FAILURE
        if not os.path.isdir(execution_dir):
            if self._tool_specification.tooltype.lower() == "executable":
                try:
                    os.makedirs(execution_dir, exist_ok=True)
                except OSError:
                    self._logger.msg_error.emit(
                        f"Creating work directory {execution_dir} failed. Try execution "
                        f"in source directory instead."
                    )
                    return ItemExecutionFinishState.FAILURE
            else:
                self._logger.msg_warning.emit(
                    "Project is not self-contained. Please make sure all Tool specification "
                    "files are in the project directory."
                )
                return ItemExecutionFinishState.FAILURE
        self._tool_instance = self._tool_specification.create_tool_instance(
            execution_dir, self._kill_completed_processes, self._logger, self
        )
        resources = forward_resources + backward_resources
        with ExitStack() as stack:
            labelled_args = labelled_resource_args(resources, stack, db_checkin=True, db_checkout=True)
            expanded_args = expand_cmd_line_args(self._cmd_line_args, labelled_args, self._logger)
            try:
                self._tool_instance.prepare(expanded_args)
            except RuntimeError as error:
                if str(error):
                    self._logger.msg_error.emit(f"Failed to prepare Tool instance: {error}")
                return ItemExecutionFinishState.FAILURE
            return_code = self._tool_instance.execute()
        self._handle_output_files(return_code, self._filter_id, forward_resources, execution_dir)
        self._tool_instance = None
        # TODO: Check what return code is 'stopped' and return ItemExecutionFinishState.STOPPED in this case
        return ItemExecutionFinishState.SUCCESS if return_code == 0 else ItemExecutionFinishState.FAILURE

    def exclude_execution(self, forward_resources, backward_resources, lock):
        """See base class."""
        super().exclude_execution(forward_resources, backward_resources, lock)
        if self._filter_id:
            self._output_dir = os.path.join(
                self._output_dir, generate_filter_subdirectory_name(forward_resources, self.hash_filter_id())
            )

    def _find_input_files(self, resources):
        """Iterates required input  files in tool specification and looks for them in the given resources.

        Args:
            resources (list): resources available

        Returns:
            Dictionary mapping required files to path where they are found, or to None if not found
        """
        file_paths = {}
        for required_path in self._tool_specification.inputfiles:
            _, filename = os.path.split(required_path)
            if not filename:
                # It's a directory
                continue
            file_paths[required_path] = find_file(filename, resources)
        return file_paths

    def _find_optional_input_files(self, resources):
        """Tries to find optional input files from previous project items in the DAG.

        Args:
            resources (list): resources available

        Returns:
            dict: Dictionary of optional input file paths or an empty dictionary if no files found. Key is the
                optional input item and value is a list of paths that matches the item.
        """
        file_paths = {}
        paths_in_resources = file_paths_from_resources(resources)
        for file_path in self._tool_specification.inputfiles_opt:
            _, pattern = os.path.split(file_path)
            if not pattern:
                # It's a directory -> skip
                continue
            found_files = _find_files_in_pattern(pattern, paths_in_resources)
            if not found_files:
                self._logger.msg_warning.emit(f"\tNo files matching pattern <b>{pattern}</b> found")
            else:
                file_paths[file_path] = found_files
        return file_paths

    def _handle_output_files(self, return_code, filter_id, forward_resources, execution_dir):
        """Copies Tool specification output files from work directory to result directory.

        Args:
            return_code (int): Tool specification process return value
            filter_id (str): filter identifier
            forward_resources (Iterable of ProjectItemResource): forward resources
            execution_dir (str): path to the execution directory
        """
        if self._tool_specification.outputfiles:
            if filter_id:
                filter_output_dir = os.path.join(
                    self._output_dir, generate_filter_subdirectory_name(forward_resources, self.hash_filter_id())
                )
                try:
                    os.makedirs(filter_output_dir, exist_ok=True)
                    self._output_dir = filter_output_dir
                except OSError:
                    self._logger.msg_error.emit(f"[OSError] Creating directory <b>{filter_output_dir}</b> failed.")
            output_dir_timestamp = _create_output_dir_timestamp()  # Get timestamp when tool finished
            # Create an output folder with timestamp and copy output directly there
            if return_code != 0:
                result_path = os.path.abspath(os.path.join(self._output_dir, "failed", output_dir_timestamp))
            else:
                result_path = os.path.abspath(os.path.join(self._output_dir, output_dir_timestamp))
            try:
                os.makedirs(result_path, exist_ok=True)
            except OSError:
                self._logger.msg_error.emit(
                    "\tError creating timestamped output directory. "
                    "Tool specification output files not copied. Please check directory permissions."
                )
                return
            # Make link to output folder
            result_anchor = (
                f"<a style='color:#BB99FF;' title='{result_path}' href='file:///{result_path}'>results directory</a>"
            )
            if filter_id:
                write_filter_id_file(filter_id, os.path.dirname(result_path))
            self._logger.msg.emit(f"*** Archiving output files to {result_anchor} ***")
            saved_files, failed_files = self._copy_output_files(result_path, execution_dir)
            if not saved_files:
                # If no files were saved
                self._logger.msg_error.emit("\tNo files saved")
            else:
                # If there are saved files
                # Split list into filenames and their paths
                filenames, _ = zip(*saved_files)
                self._logger.msg.emit(f"\tCopied <b>{len(filenames)}</b> file(s) to results directory")
            if failed_files:
                # If saving some or all files failed
                f_str = "\n".join(failed_files)
                failed_files_anchor = (
                    f"<a style='color:#ff0000;' title='{f_str}' href='#'>Some output files were not found</a>"
                )
                self._logger.msg.emit(f"\t<b>{failed_files_anchor}")
        else:
            tip_anchor = (
                "<a style='color:#99CCFF;' title='By adding <b>output files</b> to the Tool specification,\n "
                "they will be archived into the results directory. In addition, <b>output files</b> will be \n"
                "passed to subsequent project items.' href='#'>Tip</a>"
            )
            self._logger.msg_warning.emit(f"No output files defined for this Tool specification. {tip_anchor}")

    def _optional_output_destination_paths(self, paths, execution_dir):
        """Returns a dictionary telling where optional output files should be copied to before execution.

        Args:
            paths (dict): key is the optional file name pattern, value is a list of paths to source files
            execution_dir (str): a path to the execution directory

        Returns:
            dict: a map from source path to destination path
        """
        destination_paths = {}
        for dst, src_paths in paths.items():
            for src_path in src_paths:
                if not os.path.exists(src_path):
                    self._logger.msg_error.emit(f"\tFile <b>{src_path}</b> does not exist")
                    continue
                # Get file name that matched the search pattern
                _, dst_fname = os.path.split(src_path)
                # Check if the search pattern included subdirectories (e.g. 'input/*.csv')
                # This means that /input/ directory should be created to work (or source) directory
                # before copying the files
                dst_subdir, _search_pattern = os.path.split(dst)
                if not dst_subdir:
                    # No subdirectories to create
                    dst_path = os.path.abspath(os.path.join(execution_dir, dst_fname))
                else:
                    # Create subdirectory structure to work or source directory
                    work_subdir_path = os.path.abspath(os.path.join(execution_dir, dst_subdir))
                    if not os.path.exists(work_subdir_path):
                        try:
                            os.makedirs(work_subdir_path, exist_ok=True)
                        except OSError:
                            self._logger.msg_error.emit(
                                f"[OSError] Creating directory <b>{work_subdir_path}</b> failed."
                            )
                            continue
                    dst_path = os.path.abspath(os.path.join(work_subdir_path, dst_fname))
                destination_paths[src_path] = dst_path
        return destination_paths

    def _output_resources_forward(self):
        """See base class"""
        return scan_for_resources(self, self._tool_specification, self._output_dir)

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        execute_in_work = item_dict["execute_in_work"]
        if execute_in_work:
            work_dir = app_settings.value("appSettings/workDir")
            if not work_dir:
                logger.msg_error.emit(f"Error: Work directory not set for project item {name}")
                work_dir = None
        else:
            work_dir = None
        specification_name = item_dict["specification"]
        specification = DBWriterExecutableItemBase._get_specification(
            name, ItemInfo.item_type(), specification_name, specifications, logger
        )
        cmd_line_args = [make_cmd_line_arg(arg) for arg in item_dict["cmd_line_args"]]
        options = item_dict.get("options", {})
        kill_completed_processes = item_dict.get("kill_completed_processes", False)
        log_process_output = item_dict.get("log_process_output", False)
        group_id = item_dict.get("group_id")
        return cls(
            name,
            work_dir,
            specification,
            cmd_line_args,
            options,
            kill_completed_processes,
            log_process_output,
            group_id,
            project_dir,
            logger,
        )


def _count_files_and_dirs(paths):
    """Counts the number of files and directories in given paths.

    Args:
        paths (list): list of paths

    Returns:
        tuple: Tuple containing the number of required files and directories.
    """
    n_dir = 0
    n_file = 0
    for path in paths:
        _, filename = os.path.split(path)
        if not filename:
            n_dir += 1
        else:
            n_file += 1
    return n_dir, n_file


def _create_output_dir_timestamp():
    """Creates a new timestamp string that is used as Tool output
    directory.

    Returns:
        Timestamp string or empty string if failed.
    """
    try:
        # Create timestamp
        stamp = datetime.datetime.fromtimestamp(time.time())
    except OverflowError:
        return ""
    extension = stamp.strftime("%Y-%m-%dT%H.%M.%S")
    return extension


def _execution_directory(work_dir, tool_specification):
    """Returns the path to the execution directory.

    The returned path is a new unique path when the user
    has selected to execute the Tool in work directory.

    The returned path is the directory where the Tool spec's
    main program file is located when the user has selected
    to execute the Tool in source directory.

    The returned path is the directory where the Tool spec's
    definition file is located when the user has selected
    to execute the Tool in source directory and the Tool
    spec has no main program file (it is an Executable Tool
    spec).

    Args:
        work_dir (str): Path to work dir or None
        tool_specification (ToolSpecification): Tool spec instance

    Returns:
        str: Full path to execution dir
    """
    if work_dir is not None:
        basedir = os.path.join(work_dir, _unique_dir_name(tool_specification))
        return basedir
    if not tool_specification.path:
        return tool_specification.default_execution_dir
    return tool_specification.path


def _find_files_in_pattern(pattern, available_file_paths):
    """Returns a list of files that match the given pattern.

    Args:
        pattern (str): File pattern
        available_file_paths (list): List of available file paths from upstream items

    Returns:
        list: List of (full) paths
    """
    extended_pattern = os.path.join("*", pattern)  # Match all absolute paths.
    return fnmatch.filter(available_file_paths, extended_pattern)


def _unique_dir_name(tool_specification):
    """Builds a unique name for Tool's work directory."""
    return tool_specification.short_name + "__" + uuid.uuid4().hex + "__toolbox"


def _filter_ansi_escape(s):
    return _ANSI_ESCAPE.sub("", s)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Merger's executable item as well as support utilities."""
from contextlib import ExitStack
from spine_engine.spine_engine import ItemExecutionFinishState
from spine_engine.utils.returning_process import ReturningProcess
from ..db_writer_executable_item_base import DBWriterExecutableItemBase
from .do_work import do_work
from .item_info import ItemInfo


class ExecutableItem(DBWriterExecutableItemBase):
    def __init__(self, name, cancel_on_error, project_dir, logger):
        """
        Args:
            name (str): item's name
            cancel_on_error (bool): if True, revert changes on error and move on
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, project_dir, logger)
        self._cancel_on_error = cancel_on_error
        self._process = None

    @staticmethod
    def item_type():
        """Returns the data store executable's type identifier string."""
        return ItemInfo.item_type()

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        cancel_on_error = item_dict["cancel_on_error"]
        return cls(name, cancel_on_error, project_dir, logger)

    def execute(self, forward_resources, backward_resources, lock):
        """See base class."""
        if not super().execute(forward_resources, backward_resources, lock):
            return ItemExecutionFinishState.FAILURE
        from_resources = [r for r in forward_resources if r.type_ == "database"]
        to_resources = [r for r in backward_resources if r.type_ == "database"]
        if not from_resources or not to_resources:
            return ItemExecutionFinishState.SUCCESS
        with ExitStack() as stack:
            from_server_urls = [stack.enter_context(resource.open()) for resource in from_resources]
            to_server_urls = [stack.enter_context(resource.open()) for resource in to_resources]
            self._process = ReturningProcess(
                target=do_work,
                args=(self._cancel_on_error, self._logs_dir, from_server_urls, to_server_urls, lock, self._logger),
            )
            return_value = self._process.run_until_complete()
            self._process = None
            if return_value[0]:
                return ItemExecutionFinishState.SUCCESS
            return ItemExecutionFinishState.FAILURE

    def stop_execution(self):
        """Stops executing this DS."""
        super().stop_execution()
        if self._process is not None:
            self._process.terminate()
            self._process = None



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains base classes for items that write to db."""
from PySide6.QtCore import Slot
from spine_engine.utils.helpers import ExecutionDirection
from spinetoolbox.project_item.project_item import ProjectItem


class DBWriterItemBase(ProjectItem):
    """Base class for items that might write to a Spine DB."""

    def successor_data_stores(self):
        for name in self._project.successor_names(self.name):
            item = self._project.get_item(name)
            if item.item_type() == "Data Store":
                yield item

    @Slot(object, object)
    def handle_execution_successful(self, execution_direction):
        """Notifies Toolbox of successful database import."""
        if execution_direction != ExecutionDirection.FORWARD:
            return
        committed_db_maps = set()
        for successor in self.successor_data_stores():
            url = successor.sql_alchemy_url()
            db_map = self._toolbox.db_mngr.db_map(url)
            if db_map:
                committed_db_maps.add(db_map)
        if committed_db_maps:
            self._toolbox.db_mngr.notify_session_committed(self, *committed_db_maps)

    def _check_write_index(self):
        """Checks if write index is valid."""
        conflicting = {}
        descendants = set(self._project.descendant_names(self.name))
        for connection in self._project.outgoing_connections(self.name):
            sibling_connections = set(self._project.incoming_connections(connection.destination)) - {connection}
            conflicting.update(
                {
                    c.source: c.destination
                    for c in sibling_connections
                    if c.write_index < connection.write_index and c.source in descendants
                }
            )
        for name, dest in conflicting.items():
            self.add_notification(
                f"This item can't be a dependency to {name}, because {name} is set to write earlier to {dest}.\n"
                f"Possible fix: increase the write_index in the arrow from {name} to {dest}"
            )



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Importer's specification."""
from spine_engine.project_item.project_item_specification import ProjectItemSpecification
from .item_info import ItemInfo


class ImporterSpecification(ProjectItemSpecification):
    """Importer's specification."""

    def __init__(self, name, mapping, description=None):
        """
        Args:
            name (str): specification's name
            mapping (dict): mapping dict
            description (str, optional): specification's description
        """
        super().__init__(name, description, ItemInfo.item_type())
        self._mapping = mapping

    @property
    def mapping(self):
        return self._mapping

    def is_equivalent(self, other):
        """See base class."""
        return self.name == other.name and self.description == other.description and self.mapping == other.mapping

    def to_dict(self):
        """See base class."""
        return {
            "name": self.name,
            "item_type": ItemInfo.item_type(),
            "mapping": self.mapping,
            "description": self.description,
        }

    @staticmethod
    def from_dict(specification_dict, logger):
        """
        Restores :class:`ImporterSpecification` from a dictionary.

        Args:
            specification_dict (dict): serialized specification
            logger (LoggerInterface): a logger

        Returns:
            ImporterSpecification: deserialized specification
        """
        name = specification_dict["name"]
        description = specification_dict.get("description", None)
        mapping = specification_dict["mapping"]
        source_type = mapping["source_type"]
        mapping["source_type"] = {
            "CSVConnector": "CSVReader",
            "DataPackageConnector": "DatapackageReader",
            "ExcelConnector": "ExcelReader",
            "GdxConnector": "GDXReader",
            "JSONConnector": "JSONReader",
            "SqlAlchemyConnector": "SQLAlchemyReader",
        }.get(source_type, source_type)
        return ImporterSpecification(name, mapping, description)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################
import multiprocessing
from unittest import mock
from spine_engine.project_item.project_item_resource import file_resource
from spine_items.importer.do_work import do_work
from spinedb_api import from_database
from spinedb_api.spine_db_client import SpineDBClient
from spinedb_api.spine_db_server import closing_spine_db_server
from spinedb_api.spine_io.importers.csv_reader import CSVReader


class MaybeIdle:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class MockProcess:
    def __init__(self):
        self.maybe_idle = MaybeIdle()


class TestDoWork:
    def test_fixed_position_mapping_with_two_files(self, tmp_path):
        log_dir = tmp_path / "log"
        log_dir.mkdir()
        file_path_1 = tmp_path / "file1.csv"
        file_path_2 = tmp_path / "file2.csv"
        with open(file_path_1, "w") as out_file:
            out_file.writelines(["data1"])
        with open(file_path_2, "w") as out_file:
            out_file.writelines(["data2"])
        process = MockProcess()
        mapping = {
            "table_mappings": {
                "data": [
                    {
                        "Mapping 1": {
                            "mapping": [
                                {"map_type": "ParameterValueList", "position": "hidden", "value": "enum"},
                                {"map_type": "ParameterValueListValue", "position": "fixed", "value": "data: 0, 0"},
                            ]
                        }
                    }
                ]
            },
            "selected_tables": ["data"],
            "table_options": {
                "data": {
                    "skip": 0,
                    "encoding": "ascii",
                    "delimiter_custom": None,
                    "quotechar": '"',
                    "has_header": False,
                }
            },
            "table_types": {
                "data": {
                    "0": "string",
                }
            },
            "table_default_column_type": {},
            "table_row_types": {},
            "source_type": "CSVReader",
        }
        resources = [file_resource("provider item", str(file_path_1)), file_resource("provider item", str(file_path_2))]
        lock = multiprocessing.RLock()
        logger = mock.MagicMock()
        with closing_spine_db_server("sqlite://") as server_url:
            result = do_work(
                process, mapping, True, "merge", str(log_dir), resources, CSVReader(None), [server_url], lock, logger
            )
            assert result == (True,)
            client = SpineDBClient.from_server_url(server_url)
            result = client.call_method("find_list_values", parameter_value_list_name="enum")
            assert len(result["result"]) == 2
            assert from_database(result["result"][0]["value"], result["result"][0]["type"]) == "data1"
            assert from_database(result["result"][1]["value"], result["result"][1]["type"]) == "data2"



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Animation class for importers and exporters."""
from PySide6.QtCore import QLineF, QObject, QPointF, QRectF, Qt, QTimeLine, Signal, Slot
from PySide6.QtGui import QColor, QFont, QFontMetrics, QPainterPath
from PySide6.QtWidgets import QGraphicsPathItem
from spinetoolbox.font import TOOLBOX_FONT
from spinetoolbox.helpers import color_from_index


class AnimationSignaller(QObject):
    animation_started = Signal()
    animation_stopped = Signal()


class ImporterExporterAnimation:
    def __init__(self, item, duration=1800, plane_count=4, pixel_size=12, loop_width=30, loop_aspect_ratio=3):
        """Initializes animation stuff.

        Args:
            item (QGraphicsItem): The item on top of which the animation should play.
        """
        self._item = item
        self._planes = []
        self._plane_count = plane_count
        self._loop_aspect_ratio = loop_aspect_ratio
        self._loop_width = loop_width
        self._loop_height = loop_width / loop_aspect_ratio
        self._font = QFont(TOOLBOX_FONT.family)
        self._font.setPixelSize(pixel_size)
        self.time_line = QTimeLine()
        self.time_line.setDuration(duration)
        self.time_line.setLoopCount(0)  # loop forever
        self.time_line.valueChanged.connect(self._handle_time_line_value_changed)
        self.time_line.stateChanged.connect(self._handle_time_line_state_changed)
        self._step = self.time_line.updateInterval() / duration

    @Slot(float)
    def _handle_time_line_value_changed(self, value):
        for plane in self._planes:
            plane.advance(value)

    @Slot(QTimeLine.State)
    def _handle_time_line_state_changed(self, new_state):
        sources = [conn.rect().center() for conn in self._item.connectors.values() if conn.incoming_links()]
        sinks = [conn.rect().center() for conn in self._item.connectors.values() if conn.outgoing_links()]
        center = self._item.rect().center()
        if not sinks:
            sinks = [center]
        source_sink_pairs = [(source, sink) for source in sources for sink in sinks]
        for source, sink in source_sink_pairs:
            loop_rect = QRectF(
                center.x() - self._loop_width / 2,
                center.y() - self._loop_height / 2,
                self._loop_width,
                self._loop_height,
            )
            # Find sweep:
            if source == sink:
                # Same source as destination: sweep half circle
                sweep = 180
            elif source.y() == sink.y():
                # Side to side: sweep full circle
                sweep = 360
            else:
                # Side to bottom: sweep 3/4 of circle
                sweep = 270
            # Find start and sense
            invert_sense = False
            if source.y() < loop_rect.bottom():
                # From the side: start at bottom
                start = 270
                # Invert sense if from the right
                invert_sense = source.x() > center.x()
            else:
                # From the bottom
                if sink.x() < center.x():
                    # To the left: start at left and invert sense
                    start = 180
                    invert_sense = True
                else:
                    # To the right: start at right
                    start = 0
            if invert_sense:
                sweep = -sweep
            middle_point = _point_at_angle(loop_rect, start)
            path = QPainterPath()
            path.moveTo(source)
            path.quadTo(_nice_ctrl_point(path.currentPosition(), middle_point), middle_point)
            path.arcTo(loop_rect, start, sweep)
            path.quadTo(_nice_ctrl_point(path.currentPosition(), sink), sink)
            self._planes += [
                self._PaperPlane(self._item, self._font, path, -i / self._plane_count, self._step, loop_rect)
                for i in range(self._plane_count)
            ]
        for k, plane in enumerate(self._planes):
            plane.color = color_from_index(k, len(self._planes))
        if new_state == QTimeLine.State.NotRunning:
            for plane in self._planes:
                plane.wipe_out()
            self._planes.clear()

    @Slot()
    def start(self):
        """Starts the animation."""
        if self.time_line.state() == QTimeLine.State.Running:
            return
        self.time_line.start()

    @staticmethod
    def percent(value):
        raise NotImplementedError()

    @Slot()
    def stop(self):
        """Stops the animation"""
        self.time_line.stop()


class _PaperPlane(QGraphicsPathItem):
    def __init__(self, parent, font, trajectory_path, percent, step, loop_rect):
        super().__init__(parent)
        self._parent = parent
        self._trajectory_path = trajectory_path
        self._percent = percent
        self._step = step
        self._loop_rect = loop_rect
        self._icon_code = "\uf1d8"
        self.setAcceptedMouseButtons(Qt.MouseButton.NoButton)
        self.setZValue(self._parent.svg_item.zValue())
        path = QPainterPath()
        path.addText(0, (1 - 0.14) * QFontMetrics(font).height(), font, self._icon_code)  # 14% baseline for FA
        self.setPath(path)
        self.setTransformOriginPoint(self.boundingRect().center())
        border_pen = self.pen()
        border_pen.setWidthF(0.5)
        self.setPen(border_pen)
        self.color = QColor(Qt.GlobalColor.white)
        self.hide()

    def advance(self, phase):
        self._percent = self._percent + self._step
        if self._percent > 1:
            self._percent = -0.8
        if self._percent < 0:
            self.hide()
            return
        self.show()
        pos = self._trajectory_path.pointAtPercent(self._percent)  #
        self.setPos(pos - self.boundingRect().center())
        if pos.y() > self._loop_rect.center().y():
            self._parent.svg_item.stackBefore(self)
            scale = 1
        else:
            self.stackBefore(self._parent.svg_item)
            scale = 1 - 0.5 * (self._loop_rect.center().y() - pos.y()) / (self._loop_rect.height() / 2)
        self.setScale(scale)
        angle = self._trajectory_path.angleAtPercent(self._percent)
        self.resetTransform()
        if 90 < angle < 270:
            self.flip_vertical()
            self.setRotation(45 + angle)
        else:
            self.setRotation(45 - angle)
        h, s, _, _ = self.color.getHslF()
        self.color.setHslF(h, s, self._lightness())
        self.setBrush(self.color)

    def _lightness(self):
        raise NotImplementedError()

    def flip_vertical(self):
        transform = self.transform()
        m11 = transform.m11()
        m12 = transform.m12()
        m13 = transform.m13()
        m21 = transform.m21()
        m22 = transform.m22()
        m23 = transform.m23()
        m31 = transform.m31()
        m32 = super().boundingRect().height() * m22
        m33 = transform.m33()
        m22 = -m22
        transform.setMatrix(m11, m12, m13, m21, m22, m23, m31, m32, m33)
        self.setTransform(transform)

    def wipe_out(self):
        scene = self.scene()
        if scene is None:
            return
        scene.removeItem(self)


def _point_at_angle(rect, angle):
    if angle == 0:
        return QPointF(rect.right(), rect.center().y())
    if angle == 180:
        return QPointF(rect.left(), rect.center().y())
    if angle == 270:
        return QPointF(rect.center().x(), rect.bottom())


def _nice_ctrl_point(p1, p2):
    if p1.x() > p2.x():
        p1, p2 = p2, p1
    line = QLineF(p1, p2)
    p = line.center()
    normal_angle = line.normalVector().angle()
    normal_line = QLineF()
    normal_line.setP1(p)
    normal_line.setAngle(normal_angle)
    normal_line.setLength(3)
    return normal_line.p2()


class _ImporterPaperPlane(_PaperPlane):
    def _lightness(self):
        threshold = 0.66
        max_darkness = 0.5
        darkness = max(0, max_darkness * (self._percent - threshold) / (1 - threshold))
        return 1 - darkness


class _ExporterPaperPlane(_PaperPlane):
    def _lightness(self):
        threshold = 0.66
        max_darkness = 0.5
        darkness = max(0, max_darkness * (threshold - self._percent) / threshold)
        return 1 - darkness


class ImporterAnimation(ImporterExporterAnimation):
    _PaperPlane = _ImporterPaperPlane


class ExporterAnimation(ImporterExporterAnimation):
    _PaperPlane = _ExporterPaperPlane



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains model for a single export preview table."""
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt


class PreviewTableModel(QAbstractTableModel):
    MAX_COLUMNS = 50

    def __init__(self):
        super().__init__()
        self._table = []
        self._mapping_name = None
        self._table_name = None
        self._row_to_map_color = {}
        self._column_to_map_color = {}
        self._max_mapping_row = 0
        self._max_mapping_column = 0
        self._in_pivot_color = None

    def clear(self):
        """Empties the table."""
        if not self._table:
            return
        self.beginResetModel()
        self._table.clear()
        self.endResetModel()

    def columnCount(self, parent=QModelIndex()):
        if not self._table:
            return 0
        return min(max(len(row) for row in self._table), self.MAX_COLUMNS)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            row = self._table[index.row()]
            column = index.column()
            if column < len(row):
                return row[column]
        if role == Qt.ItemDataRole.BackgroundRole:
            if self._index_in_pivot(index):
                return self._in_pivot_color
            color_row = self._row_to_map_color.get(index.row())
            if color_row is not None and index.column() > self._max_mapping_column:
                return color_row
            color_column = self._column_to_map_color.get(index.column())
            if color_column is not None and index.row() > self._max_mapping_row:
                return color_column
        return None

    def _index_in_pivot(self, index):
        return index.row() > self._max_mapping_row and index.column() > self._max_mapping_column

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            return section + 1
        return None

    def mapping_name(self):
        """Returns current mapping's name.

        Returns:
            str: mapping name
        """
        return self._mapping_name

    def reset(self, mapping_name, table_name, table, mapping_colors):
        """
        Resets model's data.

        Args:
            mapping_name (str): current mapping's name
            table_name (str): current table's name
            table (list): table data
            mapping_colors (dict): mapping int position to QColor
        """
        self.beginResetModel()
        self._mapping_name = mapping_name
        self._table_name = table_name
        self._table = table
        self._reset_colors(mapping_colors)
        self.endResetModel()

    def _reset_colors(self, mapping_colors):
        """Updates attributes for coloring cells.

        Args:
            mapping_colors (dict): mapping int position to QColor
        """
        self._row_to_map_color = {}
        self._column_to_map_color = {}
        # Compute in-pivot color
        positions = list(mapping_colors)
        is_pivoted = any(p < 0 for p in positions[:-1])
        if is_pivoted:
            p = positions.pop(-1)
            self._in_pivot_color = mapping_colors[p]
        # Compute row and column colors
        for position in positions:
            color = mapping_colors[position]
            if position < 0:
                row = -(position + 1)
                self._row_to_map_color[row] = color
            else:
                column = position
                self._column_to_map_color[column] = color
        # Compact the rows
        sorted_row_to_map_color = {row: self._row_to_map_color[row] for row in sorted(self._row_to_map_color)}
        self._row_to_map_color = dict(enumerate(sorted_row_to_map_color.values()))
        # Compute maxes
        self._max_mapping_row = max(self._row_to_map_color, default=-1)
        self._max_mapping_column = max(self._column_to_map_color, default=-1)

    def rowCount(self, parent=QModelIndex()):
        return len(self._table)

    def table_name(self):
        """Returns current table's name.

        Returns:
            str: table name
        """
        return self._table_name
